import hashlib
import pickle
import time
from functools import wraps


class RateLimitException(Exception):
    pass


class RateLimit(object):
    """
    Rate limit implementation. Allows up to `number` of events every `per`
    seconds.
    """
    def __init__(self, database, name, limit=5, per=60, debug=False):
        """
        :param database: :py:class:`Database` instance.
        :param name: Namespace for this cache.
        :param int limit: Number of events allowed during a given time period.
        :param int per: Time period the ``limit` applies to, in seconds.
        :param debug: Disable rate-limit for debugging purposes. All events
                      will appear to be allowed and valid.
        """
        self.database = database
        self.name = name
        self._limit = limit
        self._per = per
        self._debug = debug

    def limit(self, key):
        """
        Function to log an event with the given key. If the ``key`` has not
        exceeded their alotted events, then the function returns ``False`` to
        indicate that no limit is being imposed.

        If the ``key`` has exceeded the number of events, then the function
        returns ``True`` indicating rate-limiting should occur.

        :param str key: A key identifying the source of the event.
        :returns: Boolean indicating whether the event should be rate-limited
            or not.
        """
        if self._debug:
            return False

        counter = self.database.List(self.name + ':' + key)
        n = len(counter)
        is_limited = False
        if n < self._limit:
            counter.prepend(str(time.time()))
        else:
            oldest = float(counter[-1])
            if time.time() - oldest < self._per:
                is_limited = True
            else:
                counter.prepend(str(time.time()))
            del counter[:self._limit]
        counter.pexpire(int(self._per * 2000))
        return is_limited

    def rate_limited(self, key_function=None):
        """
        Function or method decorator that will prevent calls to the decorated
        function when the number of events has been exceeded for the given
        time period.

        It is probably important that you take care to choose an appropiate
        key function. For instance, if rate-limiting a web-page you might use
        the requesting user's IP as the key.

        If the number of allowed events has been exceedd, a
        ``RateLimitException`` will be raised.

        :param key_function: Function that accepts the params of the decorated
            function and returns a string key. If not provided, a hash of the
            args and kwargs will be used.
        :returns: If the call is not rate-limited, then the return value will
            be that of the decorated function.
        :raises: ``RateLimitException``.
        """
        if key_function is None:
            def key_function(*args, **kwargs):
                data = pickle.dumps((args, sorted(kwargs.items())))
                return hashlib.md5(data).hexdigest()

        def decorator(fn):
            @wraps(fn)
            def inner(*args, **kwargs):
                key = key_function(*args, **kwargs)
                if self.limit(key):
                    raise RateLimitException(
                        'Call to %s exceeded %s events in %s seconds.' % (
                            fn.__name__, self._limit, self._per))
                return fn(*args, **kwargs)
            return inner
        return decorator



import re

from walrus.search.metaphone import dm as double_metaphone
from walrus.search.porter import PorterStemmer
from walrus.utils import decode
from walrus.utils import load_stopwords


class Tokenizer(object):
    def __init__(self, stemmer=True, metaphone=False,
                 stopwords_file='stopwords.txt', min_word_length=None):
        self._use_stemmer = stemmer
        self._use_metaphone = metaphone
        self._min_word_length = min_word_length
        self._symbols_re = re.compile(
            '[\.,;:"\'\\/!@#\$%\?\*\(\)\-=+\[\]\{\}_]')

        self._stopwords = self._load_stopwords(stopwords_file)

    def _load_stopwords(self, filename):
        stopwords = load_stopwords(filename)
        if stopwords:
            return set(stopwords.splitlines())

    def split_phrase(self, phrase):
        """Split the document or search query into tokens."""
        return self._symbols_re.sub(' ', phrase).split()

    def stem(self, words):
        """
        Use the porter stemmer to generate consistent forms of
        words, e.g.::

            from walrus.search.utils import PorterStemmer
            stemmer = PorterStemmer()
            for word in ['faith', 'faiths', 'faithful']:
                print s.stem(word, 0, len(word) - 1)

            # Prints:
            # faith
            # faith
            # faith
        """
        stemmer = PorterStemmer()
        _stem = stemmer.stem
        for word in words:
            yield _stem(word, 0, len(word) - 1)

    def metaphone(self, words):
        """
        Apply the double metaphone algorithm to the given words.
        Using metaphone allows the search index to tolerate
        misspellings and small typos.

        Example::

            >>> from walrus.search.metaphone import dm as metaphone
            >>> print metaphone('walrus')
            ('ALRS', 'FLRS')

            >>> print metaphone('python')
            ('P0N', 'PTN')

            >>> print metaphone('pithonn')
            ('P0N', 'PTN')
        """
        for word in words:
            r = 0
            for w in double_metaphone(word):
                if w:
                    w = w.strip()
                    if w:
                        r += 1
                        yield w
            if not r:
                yield word

    def tokenize(self, value):
        """
        Split the incoming value into tokens and process each token,
        optionally stemming or running metaphone.

        :returns: A ``dict`` mapping token to score. The score is
            based on the relative frequency of the word in the
            document.
        """
        words = self.split_phrase(decode(value).lower())
        if self._stopwords:
            words = [w for w in words if w not in self._stopwords]
        if self._min_word_length:
            words = [w for w in words if len(w) >= self._min_word_length]

        fraction = 1. / (len(words) + 1)  # Prevent division by zero.

        # Apply optional transformations.
        if self._use_stemmer:
            words = self.stem(words)
        if self._use_metaphone:
            words = self.metaphone(words)

        scores = {}
        for word in words:
            scores.setdefault(word, 0)
            scores[word] += fraction
        return scores



from functools import wraps
import os


class Lock(object):
    """
    Lock implementation. Can also be used as a context-manager or
    decorator.

    Unlike the redis-py lock implementation, this Lock does not
    use a spin-loop when blocking to acquire the lock. Instead,
    it performs a blocking pop on a list. When a lock is released,
    a value is pushed into this list, signalling that the lock is
    available.

    The lock uses Lua scripts to ensure the atomicity of its
    operations.

    You can set a TTL on a lock to reduce the potential for deadlocks
    in the event of a crash. If a lock is not released before it
    exceeds its TTL, and threads that are blocked waiting for the
    lock could potentially re-acquire it.

    .. note:: TTL is specified in **milliseconds**.

    Locks can be used as context managers or as decorators:

    .. code-block:: python

        lock = db.lock('my-lock')

        with lock:
            perform_some_calculations()

        @lock
        def another_function():
            # The lock will be acquired when this function is
            # called, and released when the function returns.
            do_some_more_calculations()
    """
    def __init__(self, database, name, ttl=None, lock_id=None):
        """
        :param database: A walrus ``Database`` instance.
        :param str name: The name for the lock.
        :param int ttl: The time-to-live for the lock in milliseconds.
        :param str lock_id: Unique identifier for the lock instance.
        """
        self.database = database
        self.name = name
        self.ttl = ttl or 0
        self._lock_id = lock_id or os.urandom(32)

    @property
    def key(self):
        return 'lock:%s' % (self.name)

    @property
    def event(self):
        return 'lock.event:%s' % (self.name)

    def acquire(self, block=True):
        """
        Acquire the lock. The lock will be held until it is released
        by calling :py:meth:`Lock.release`. If the lock was
        initialized with a ``ttl``, then the lock will be released
        automatically after the given number of milliseconds.

        By default this method will block until the lock becomes
        free (either by being released or expiring). The blocking is
        accomplished by performing a blocking left-pop on a list, as
        opposed to a spin-loop.

        If you specify ``block=False``, then the method will return
        ``False`` if the lock could not be acquired.

        :param bool block: Whether to block while waiting to acquire
            the lock.
        :returns: Returns ``True`` if the lock was acquired.
        """
        while True:
            acquired = self.database.run_script(
                'lock_acquire',
                keys=[self.key],
                args=[self._lock_id, self.ttl])
            if acquired == 1 or not block:
                return acquired == 1

            # Perform a blocking pop on the event key. When a lock
            # is released, a value is pushed into the list, which
            # signals listeners that the lock is available.
            self.database.blpop(self.event, self.ttl)

    def release(self):
        """
        Release the lock.

        :returns: Returns ``True`` if the lock was released.
        """
        unlocked = self.database.run_script(
            'lock_release',
            keys=[self.key, self.event],
            args=[self._lock_id])
        return unlocked != 0

    def clear(self):
        """
        Clear the lock, allowing it to be acquired. Do not use this
        method except to recover from a deadlock. Otherwise you should
        use :py:meth:`Lock.release`.
        """
        self.database.delete(self.key)
        self.database.delete(self.event)

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.release():
            raise RuntimeError('Error releasing lock "%s".' % self.name)

    def __call__(self, fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)
        return inner



import json
import re

from walrus.utils import decode
from walrus.utils import encode
from walrus.utils import load_stopwords
from walrus.utils import PY3


class Autocomplete(object):
    """
    Autocompletion for ascii-encoded string data. Titles are stored,
    along with any corollary data in Redis. Substrings of the title
    are stored in sorted sets using a unique scoring algorithm. The
    scoring algorithm aims to return results in a sensible order,
    by looking at the entire title and the position of the matched
    substring within the title.

    Additionally, the autocomplete object supports boosting search
    results by object ID or object type.
    """
    def __init__(self, database, namespace='walrus', cache_timeout=600,
                 stopwords_file='stopwords.txt', use_json=True):
        """
        :param database: A :py:class:`Database` instance.
        :param namespace: Namespace to prefix keys used to store
            metadata.
        :param cache_timeout: Complex searches using boosts will be
            cached. Specify the amount of time these results are
            cached for.
        :param stopwords_file: Filename containing newline-separated
            stopwords.
        :param bool use_json: Whether object data should be
            serialized as JSON.
        """
        self.database = database
        self.namespace = namespace
        self._cache_timeout = cache_timeout
        self._stopwords_file = stopwords_file
        self._use_json = use_json
        self._load_stopwords()

        self._data = self.database.Hash('%s:d' % self.namespace)
        self._title_data = self.database.Hash('%s:t' % self.namespace)
        self._boosts = self.database.Hash('%s:b' % self.namespace)

        self._max_title = 10
        self._offset = self.score_token('z' * self._max_title) + 1

    def _load_stopwords(self):
        if self._stopwords_file:
            stopwords = load_stopwords(self._stopwords_file)
            self._stopwords = set(stopwords.splitlines())
        else:
            self._stopwords = set()

    def tokenize_title(self, phrase, stopwords=True):
        if isinstance(phrase, bytes):
            phrase = phrase.decode("utf-8")
        phrase = re.sub('[^a-z0-9_\-\s]', '', phrase.lower())
        if stopwords:
            return [w for w in phrase.split() if w not in self._stopwords]
        else:
            return phrase.split()

    def score_token(self, token):
        l = len(token)
        a = ord('a') - 2
        score = 0

        for i in range(self._max_title):
            if i < l:
                c = ord(token[i]) - a
                if c < 2 or c > 27:
                    c = 1
            else:
                c = 1
            score += c * (27 ** (self._max_title - i - 1))

        return score

    def substrings(self, w):
        for i in range(1, len(w)):
            yield w[:i]
        yield w

    def object_key(self, obj_id, obj_type):
        return '%s\x01%s' % (obj_id, obj_type or '')

    def word_key(self, word):
        return '%s:s:%s' % (self.namespace, word)

    def store(self, obj_id, title=None, data=None, obj_type=None):
        """
        Store data in the autocomplete index.

        :param obj_id: Either a unique identifier for the object
            being indexed or the word/phrase to be indexed.
        :param title: The word or phrase to be indexed. If not
            provided, the ``obj_id`` will be used as the title.
        :param data: Arbitrary data to index, which will be
            returned when searching for results. If not provided,
            this value will default to the title being indexed.
        :param obj_type: Optional object type. Since results can be
            boosted by type, you might find it useful to specify this
            when storing multiple types of objects.

        You have the option of storing several types of data as
        defined by the parameters. At the minimum, you can specify
        an ``obj_id``, which will be the word or phrase you wish to
        index. Alternatively, if for instance you were indexing blog
        posts, you might specify all parameters.
        """
        if title is None:
            title = obj_id
        if data is None:
            data = title
        obj_type = obj_type or ''

        if self._use_json:
            data = json.dumps(data)

        combined_id = self.object_key(obj_id, obj_type)

        if self.exists(obj_id, obj_type):
            stored_title = self._title_data[combined_id]
            if stored_title == title:
                self._data[combined_id] = data
                return
            else:
                self.remove(obj_id, obj_type)

        self._data[combined_id] = data
        self._title_data[combined_id] = title

        clean_title = ' '.join(self.tokenize_title(title))
        title_score = self.score_token(clean_title)

        for idx, word in enumerate(self.tokenize_title(title)):
            word_score = self.score_token(word)
            position_score = word_score + (self._offset * idx)
            key_score = position_score + title_score
            for substring in self.substrings(word):
                self.database.zadd(
                    self.word_key(substring),
                    combined_id,
                    key_score)

        return True

    def remove(self, obj_id, obj_type=None):
        """
        Remove an object identified by the given ``obj_id`` (and
        optionally ``obj_type``) from the search index.

        :param obj_id: The object's unique identifier.
        :param obj_type: The object's type.
        """
        if not self.exists(obj_id, obj_type):
            raise KeyError('Object not found.')

        combined_id = self.object_key(obj_id, obj_type)
        title = self._title_data[combined_id]

        for word in self.tokenize_title(title):
            for substring in self.substrings(word):
                key = self.word_key(substring)
                if not self.database.zrange(key, 1, 2):
                    self.database.delete(key)
                else:
                    self.database.zrem(key, combined_id)

        del self._data[combined_id]
        del self._title_data[combined_id]
        del self._boosts[combined_id]

    def exists(self, obj_id, obj_type=None):
        """
        Return whether the given object exists in the search index.

        :param obj_id: The object's unique identifier.
        :param obj_type: The object's type.
        """
        return self.object_key(obj_id, obj_type) in self._data

    def boost_object(self, obj_id=None, obj_type=None, multiplier=1.1,
                     relative=True):
        """
        Boost search results for the given object or type by the
        amount specified. When the ``multiplier`` is greater than
        1, the results will percolate to the top. Values between
        0 and 1 will percolate results to the bottom.

        Either an ``obj_id`` or ``obj_type`` (or both) must be
        specified.

        :param obj_id: An object's unique identifier (optional).
        :param obj_type: The object's type (optional).
        :param multiplier: A positive floating-point number.
        :param relative: If ``True``, then any pre-existing saved
            boost will be updated using the given multiplier.

        Examples:

        .. code-block:: python

            # Make all objects of type=photos percolate to top.
            ac.boost_object(obj_type='photo', multiplier=2.0)

            # Boost a particularly popular blog entry.
            ac.boost_object(
                popular_entry.id,
                'entry',
                multipler=5.0,
                relative=False)
        """
        combined_id = self.object_key(obj_id or '', obj_type or '')
        if relative:
            current = float(self._boosts[combined_id] or 1.0)
            self._boosts[combined_id] = current * multiplier
        else:
            self._boosts[combined_id] = multiplier

    def _load_objects(self, obj_id_zset, limit, chunk_size=1000):
        ct = i = 0
        while True:
            id_chunk = obj_id_zset[i:i + chunk_size]
            if not id_chunk:
                raise StopIteration

            i += chunk_size
            for raw_data in self._data[id_chunk]:
                if not raw_data:
                    continue
                if self._use_json:
                    yield json.loads(raw_data.decode('utf-8'))
                else:
                    yield raw_data
                ct += 1
                if limit and ct == limit:
                    raise StopIteration

    def _load_saved_boosts(self):
        boosts = {}
        for combined_id, score in self._boosts:
            obj_id, obj_type = combined_id.split(encode('\x01'), 1)
            score = float(score)
            if obj_id and obj_type:
                boosts[combined_id] = score
            elif obj_id:
                boosts[obj_id] = score
            elif obj_type:
                boosts[obj_type] = score
        return boosts

    def search(self, phrase, limit=None, boosts=None, chunk_size=1000):
        """
        Perform a search for the given phrase. Objects whose title
        matches the search will be returned. The values returned
        will be whatever you specified as the ``data`` parameter
        when you called :py:meth:`~Autocomplete.store`.

        :param phrase: One or more words or substrings.
        :param int limit: Limit size of the result set.
        :param dict boosts: A mapping of object id/object type to
            floating point multipliers.
        :returns: A list containing the object data for objects
            matching the search phrase.
        """
        cleaned = self.tokenize_title(phrase, stopwords=False)
        if not cleaned:
            raise StopIteration

        all_boosts = self._load_saved_boosts()
        if PY3 and boosts:
            for key in boosts:
                all_boosts[encode(key)] = boosts[key]
        elif boosts:
            all_boosts.update(boosts)

        if len(cleaned) == 1 and not all_boosts:
            result_key = self.word_key(cleaned[0])
        else:
            result_key = self.get_cache_key(cleaned, all_boosts)
            if result_key not in self.database:
                self.database.zinterstore(
                    result_key,
                    list(map(self.word_key, cleaned)))
            self.database.expire(result_key, self._cache_timeout)

        results = self.database.ZSet(result_key)
        if all_boosts:
            for raw_id, score in results[0:0, True]:
                orig_score = score
                for identifier in raw_id.split(encode('\x01'), 1):
                    if identifier and identifier in all_boosts:
                        score *= 1 / all_boosts[identifier]

                if orig_score != score:
                    results[raw_id] = score

        for result in self._load_objects(results, limit, chunk_size):
            yield result

    def get_cache_key(self, phrases, boosts):
        if boosts:
            boost_key = '|'.join(sorted(
                '%s:%s' % (k, v) for k, v in boosts.items()))
        else:
            boost_key = ''
        phrase_key = '|'.join(phrases)
        return '%s:c:%s:%s' % (self.namespace, phrase_key, boost_key)

    def list_data(self):
        """
        Return all the data stored in the autocomplete index. If the data was
        stored as serialized JSON, then it will be de-serialized before being
        returned.

        :rtype: list
        """
        values = self._data.values()
        return [json.loads(v) for v in values] if self._use_json else values

    def list_titles(self):
        """
        Return the titles of all objects stored in the autocomplete index.

        :rtype: list
        """
        return self._title_data.values()

    def flush(self, batch_size=1000):
        """
        Delete all autocomplete indexes and metadata.
        """
        keys = self.database.keys(self.namespace + ':*')
        for i in range(0, len(keys), batch_size):
            self.database.delete(*keys[i:i + batch_size])



import os
import sys


PY3 = sys.version_info[0] == 3

if PY3:
    unicode_type = str
    basestring_type = (str, bytes)
else:
    unicode_type = unicode
    basestring_type = basestring


def encode(s):
    return s.encode('utf-8') if isinstance(s, unicode_type) else s


def decode(s):
    return s.decode('utf-8') if isinstance(s, bytes) else s


class memoize(dict):
    def __init__(self, fn):
        self._fn = fn

    def __call__(self, *args):
        return self[args]

    def __missing__(self, key):
        result = self[key] = self._fn(*key)
        return result


@memoize
def load_stopwords(stopwords_file):
    path, filename = os.path.split(stopwords_file)
    if not path:
        path = os.path.dirname(__file__)
    filename = os.path.join(path, filename)
    if not os.path.exists(filename):
        return

    with open(filename) as fh:
        return fh.read()



from copy import deepcopy
import datetime
import json
import pickle
import re
import sys
import time
import uuid
from warnings import warn

from walrus.containers import Array
from walrus.containers import Hash
from walrus.containers import HyperLogLog
from walrus.containers import List
from walrus.containers import Set
from walrus.containers import ZSet
from walrus.query import ABSOLUTE
from walrus.query import CONTINUOUS
from walrus.query import Desc
from walrus.query import Executor
from walrus.query import FTS
from walrus.query import Node
from walrus.search import Tokenizer
from walrus.utils import basestring_type
from walrus.utils import decode
from walrus.utils import encode
from walrus.utils import PY3
from walrus.utils import unicode_type


class Field(Node):
    """
    Named attribute on a model that will hold a value of the given
    type. Fields are declared as attributes on a model class.

    Example::

        walrus_db = Database()

        class User(Model):
            __database__ = walrus_db
            __namespace__ = 'my-app'

            # Use the user's email address as the primary key.
            # All primary key fields will also get a secondary
            # index, so there's no need to specify index=True.
            email = TextField(primary_key=True)

            # Store the user's interests in a free-form text
            # field. Also create a secondary full-text search
            # index on this field.
            interests = TextField(
                fts=True,
                stemmer=True,
                min_word_length=3)

        class Note(Model):
            __database__ = walrus_app
            __namespace__ = 'my-app'

            # A note is associated with a user. We will create a
            # secondary index on this field so we can efficiently
            # retrieve all notes created by a specific user.
            user_email = TextField(index=True)

            # Store the note content in a searchable text field. Use
            # the double-metaphone algorithm to index the content.
            content = TextField(
                fts=True,
                stemmer=True,
                metaphone=True)

            # Store the timestamp the note was created automatically.
            # Note that we do not call `now()`, but rather pass the
            # function itself.
            timestamp = DateTimeField(default=datetime.datetime.now)
    """
    _coerce = None

    def __init__(self, index=False, primary_key=False, default=None):
        """
        :param bool index: Use this field as an index. Indexed
            fields will support :py:meth:`Model.get` lookups.
        :param bool primary_key: Use this field as the primary key.
        """
        self._index = index or primary_key
        self._primary_key = primary_key
        self._default = default

    def _generate_key(self):
        raise NotImplementedError

    def db_value(self, value):
        if self._coerce:
            return self._coerce(value)
        return value

    def python_value(self, value):
        if self._coerce:
            return self._coerce(value)
        return value

    def add_to_class(self, model_class, name):
        self.model_class = model_class
        self.name = name
        setattr(model_class, name, self)

    def __get__(self, instance, instance_type=None):
        if instance is not None:
            return instance._data.get(self.name)
        return self

    def __set__(self, instance, value):
        instance._data[self.name] = value

    def get_index(self, op):
        indexes = self.get_indexes()
        for index in indexes:
            if op in index.operations:
                return index

        raise ValueError('Operation %s is not supported by an index.' % op)

    def get_indexes(self):
        """
        Return a list of secondary indexes to create for the
        field. For instance, a TextField might have a full-text
        search index, whereas an IntegerField would have a scalar
        index that supported range queries.
        """
        return [AbsoluteIndex(self)]


class _ScalarField(Field):
    def get_indexes(self):
        return [AbsoluteIndex(self), ContinuousIndex(self)]


class IntegerField(_ScalarField):
    """Store integer values."""
    _coerce = int

    def db_value(self, value):
        return 0 if value is None else int(value)


class AutoIncrementField(IntegerField):
    """Auto-incrementing primary key field."""
    def __init__(self, *args, **kwargs):
        kwargs['primary_key'] = True
        return super(AutoIncrementField, self).__init__(*args, **kwargs)

    def _generate_key(self):
        query_helper = self.model_class._query
        key = query_helper.make_key(self.name, '_sequence')
        return self.model_class.__database__.incr(key)


class FloatField(_ScalarField):
    """Store floating point values."""
    _coerce = float

    def db_value(self, value):
        return 0. if value is None else float(value)


class ByteField(Field):
    """Store arbitrary bytes."""
    def db_value(self, value):
        if isinstance(value, unicode_type):
            value = value.encode('utf-8')
        elif value is None:
            value = b''
        return value


class TextField(Field):
    """
    Store unicode strings, encoded as UTF-8. :py:class:`TextField`
    also supports full-text search through the optional ``fts``
    parameter.

    .. note:: If full-text search is enabled for the field, then
        the ``index`` argument is implied.

    :param bool fts: Enable simple full-text search.
    :param bool stemmer: Use porter stemmer to process words.
    :param bool metaphone: Use the double metaphone algorithm to
        process words.
    :param str stopwords_file: File containing stopwords, one per
        line. If not specified, the default stopwords will be used.
    :param int min_word_length: Minimum length (inclusive) of word
        to be included in search index.
    """
    def __init__(self, fts=False, stemmer=True, metaphone=False,
                 stopwords_file=None, min_word_length=None, *args, **kwargs):
        super(TextField, self).__init__(*args, **kwargs)
        self._fts = fts
        self._stemmer = stemmer
        self._metaphone = metaphone
        self._stopwords_file = stopwords_file
        self._min_word_length = min_word_length
        self._index = self._index or self._fts

    def db_value(self, value):
        return b'' if value is None else encode(value)

    def python_value(self, value):
        return decode(value)

    def get_indexes(self):
        indexes = super(TextField, self).get_indexes()
        if self._fts:
            indexes.append(FullTextIndex(
                self,
                self._stemmer,
                self._metaphone,
                self._stopwords_file,
                self._min_word_length))
        return indexes


class BooleanField(Field):
    """Store boolean values."""
    def db_value(self, value):
        return '1' if value else '0'

    def python_value(self, value):
        return decode(value) == '1'


class UUIDField(Field):
    """Store unique IDs. Can be used as primary key."""
    def __init__(self, **kwargs):
        kwargs['index'] = True
        super(UUIDField, self).__init__(**kwargs)

    def db_value(self, value):
        return encode(value.hex if value is not None else '')

    def python_value(self, value):
        return uuid.UUID(decode(value)) if value else None

    def _generate_key(self):
        return uuid.uuid4()


class DateTimeField(_ScalarField):
    """Store Python datetime objects."""
    def db_value(self, value):
        if value is None:
            return b''

        timestamp = time.mktime(value.timetuple())
        micro = value.microsecond * (10 ** -6)
        return timestamp + micro

    def python_value(self, value):
        if not value:
            return None
        elif isinstance(value, (basestring_type, int, float)):
            return datetime.datetime.fromtimestamp(float(value))
        else:
            return value


class DateField(DateTimeField):
    """Store Python date objects."""
    def db_value(self, value):
        if value is None:
            return b''
        return time.mktime(value.timetuple())

    def python_value(self, value):
        if not value:
            return None
        elif isinstance(value, (basestring_type, int, float)):
            return datetime.datetime.fromtimestamp(float(value)).date()
        else:
            return value


class JSONField(Field):
    """Store arbitrary JSON data."""
    def db_value(self, value):
        return encode(json.dumps(value))

    def python_value(self, value):
        return json.loads(decode(value))


class PickledField(Field):
    """Store arbitrary Python objects."""
    def db_value(self, value):
        return pickle.dumps(value, pickle.HIGHEST_PROTOCOL)

    def python_value(self, value):
        return pickle.loads(value)


class _ContainerField(Field):
    container_class = None

    def __init__(self, *args, **kwargs):
        super(_ContainerField, self).__init__(*args, **kwargs)
        if self._primary_key:
            raise ValueError('Container fields cannot be primary keys.')
        if self._index:
            raise ValueError('Container fields cannot be indexed.')

    def _get_container(self, instance):
        return self.container_class(
            self.model_class.__database__,
            self.__key__(instance))

    def __key__(self, instance):
        return self.model_class._query.make_key(
            'container',
            self.name,
            instance.get_hash_id())

    def __get__(self, instance, instance_type=None):
        if instance is not None:
            if not instance.get_id():
                raise ValueError('Model must have a primary key before '
                                 'container attributes can be accessed.')
            return self._get_container(instance)
        return self

    def __set__(self, instance, instance_type=None):
        raise ValueError('Cannot set the value of a container field.')

    def _delete(self, instance):
        self._get_container(instance).clear()


class HashField(_ContainerField):
    """Store values in a Redis hash."""
    container_class = Hash


class ListField(_ContainerField):
    """Store values in a Redis list."""
    container_class = List


class SetField(_ContainerField):
    """Store values in a Redis set."""
    container_class = Set


class ZSetField(_ContainerField):
    """Store values in a Redis sorted set."""
    container_class = ZSet


class Query(object):
    def __init__(self, model_class):
        self.model_class = model_class

    @property
    def _base_key(self):
        model_name = self.model_class.__name__.lower()
        if self.model_class.__namespace__:
            return '%s|%s:' % (self.model_class.__namespace__, model_name)
        return '%s:' % model_name

    def make_key(self, *parts):
        """Generate a namespaced key for the given path."""
        separator = getattr(self.model_class, 'index_separator', '.')
        parts = map(decode, parts)
        return '%s%s' % (self._base_key, separator.join(map(str, parts)))

    def get_primary_hash_key(self, primary_key):
        pk_field = self.model_class._fields[self.model_class._primary_key]
        return self.make_key('id', pk_field.db_value(primary_key))

    def all_index(self):
        return self.model_class.__database__.Set(self.make_key('all'))


class BaseIndex(object):
    operations = None

    def __init__(self, field):
        self.field = field
        self.__database__ = self.field.model_class.__database__
        self.query_helper = self.field.model_class._query

    def field_value(self, instance):
        return self.field.db_value(getattr(instance, self.field.name))

    def get_key(self, instance, value):
        raise NotImplementedError

    def store_instance(self, key, instance, value):
        raise NotImplementedError

    def delete_instance(self, key, instance, value):
        raise NotImplementedError

    def save(self, instance):
        value = self.field_value(instance)
        key = self.get_key(value)
        self.store_instance(key, instance, value)

    def remove(self, instance):
        value = self.field_value(instance)
        key = self.get_key(value)
        self.delete_instance(key, instance, value)


class AbsoluteIndex(BaseIndex):
    operations = ABSOLUTE

    def get_key(self, value):
        key = self.query_helper.make_key(
            self.field.name,
            'absolute',
            value)
        return self.__database__.Set(key)

    def store_instance(self, key, instance, value):
        key.add(instance.get_hash_id())

    def delete_instance(self, key, instance, value):
        key.remove(instance.get_hash_id())
        if len(key) == 0:
            key.clear()


class ContinuousIndex(BaseIndex):
    operations = CONTINUOUS

    def get_key(self, value):
        key = self.query_helper.make_key(
            self.field.name,
            'continuous')
        return self.__database__.ZSet(key)

    def store_instance(self, key, instance, value):
        key[instance.get_hash_id()] = value

    def delete_instance(self, key, instance, value):
        del key[instance.get_hash_id()]
        if len(key) == 0:
            key.clear()


class FullTextIndex(BaseIndex):
    operations = FTS

    def __init__(self, field, stemmer=True, metaphone=False,
                 stopwords_file=None, min_word_length=None):
        super(FullTextIndex, self).__init__(field)
        self.tokenizer = Tokenizer(
            stemmer=stemmer,
            metaphone=metaphone,
            stopwords_file=stopwords_file or 'stopwords.txt',
            min_word_length=min_word_length)

    def get_key(self, value):
        key = self.query_helper.make_key(
            self.field.name,
            'fts',
            value)
        return self.__database__.ZSet(key)

    def store_instance(self, key, instance, value):
        hash_id = instance.get_hash_id()
        for word, score in self.tokenizer.tokenize(value).items():
            key = self.get_key(word)
            key[hash_id] = score

    def delete_instance(self, key, instance, value):
        hash_id = instance.get_hash_id()
        for word in self.tokenizer.tokenize(value):
            key = self.get_key(word)
            del key[hash_id]
            if len(key) == 0:
                key.clear()


class BaseModel(type):
    def __new__(cls, name, bases, attrs):
        if not bases:
            return super(BaseModel, cls).__new__(cls, name, bases, attrs)

        if 'database' in attrs:
            warn('"database" has been deprecated in favor of "__database__" '
                 'for Walrus models.', DeprecationWarning)
            attrs['__database__'] = attrs.pop('database')
        if 'namespace' in attrs:
            warn('"namespace" has been deprecated in favor of "__namespace__" '
                 'for Walrus models.', DeprecationWarning)
            attrs['__namespace__'] = attrs.pop('namespace')

        # Declarative base juju.
        ignore = set()
        primary_key = None

        for key, value in attrs.items():
            if isinstance(value, Field) and value._primary_key:
                primary_key = (key, value)

        for base in bases:
            for key, value in base.__dict__.items():
                if key in attrs:
                    continue
                if isinstance(value, Field):
                    if value._primary_key and primary_key:
                        ignore.add(key)
                    else:
                        if value._primary_key:
                            primary_key = (key, value)
                        attrs[key] = deepcopy(value)

        if not primary_key:
            attrs['_id'] = AutoIncrementField()
            primary_key = ('_id', attrs['_id'])

        model_class = super(BaseModel, cls).__new__(cls, name, bases, attrs)
        model_class._data = None

        defaults = {}
        fields = {}
        indexes = []
        for key, value in model_class.__dict__.items():
            if isinstance(value, Field) and key not in ignore:
                value.add_to_class(model_class, key)
                if value._index:
                    indexes.append(value)
                fields[key] = value
                if value._default is not None:
                    defaults[key] = value._default

        model_class._defaults = defaults
        model_class._fields = fields
        model_class._indexes = indexes
        model_class._primary_key = primary_key[0]
        model_class._query = Query(model_class)
        return model_class


def _with_metaclass(meta, base=object):
    return meta("NewBase", (base,), {'__database__': None,
                                     '__namespace__': None})


class Model(_with_metaclass(BaseModel)):
    """
    A collection of fields to be stored in the database. Walrus
    stores model instance data in hashes keyed by a combination of
    model name and primary key value. Instance attributes are
    automatically converted to values suitable for storage in Redis
    (i.e., datetime becomes timestamp), and vice-versa.

    Additionally, model fields can be ``indexed``, which allows
    filtering. There are three types of indexes:

    * Absolute
    * Scalar
    * Full-text search

    Absolute indexes are used for values like strings or UUIDs and
    support only equality and inequality checks.

    Scalar indexes are for numeric values as well as datetimes,
    and support equality, inequality, and greater or less-than.

    The final type of index, FullText, can only be used with the
    :py:class:`TextField`. FullText indexes allow search using
    the ``match()`` method. For more info, see :ref:`fts`.
    """
    #: **Required**: the :py:class:`Database` instance to use to
    #: persist model data.
    __database__ = None

    #: **Optional**: namespace to use for model data.
    __namespace__ = None

    #: **Required**: character to use as a delimiter for indexes, default "."
    index_separator = '.'

    def __init__(self, *args, **kwargs):
        self._data = {}
        self._load_default_dict()
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return '<%s: %s>' % (type(self).__name__, self.get_id())

    def _load_default_dict(self):
        for field_name, default in self._defaults.items():
            if callable(default):
                default = default()
            setattr(self, field_name, default)

    def incr(self, field, incr_by=1):
        """
        Increment the value stored in the given field by the specified
        amount. Any indexes will be updated at the time ``incr()`` is
        called.

        :param Field field: A field instance.
        :param incr_by: An ``int`` or ``float``.

        Example:

        .. code-block:: python

            # Retrieve a page counter object for the given URL.
            page_count = PageCounter.get(PageCounter.url == url)

            # Update the hit count, persisting to the database and
            # updating secondary indexes in one go.
            page_count.incr(PageCounter.hits)
        """
        model_hash = self.to_hash()

        # Remove the value from the index.
        for index in field.get_indexes():
            index.remove(self)

        if isinstance(incr_by, int):
            new_val = model_hash.incr(field.name, incr_by)
        else:
            new_val = model_hash.incr_float(field.name, incr_by)
        setattr(self, field.name, new_val)

        # Re-index the new value.
        for index in field.get_indexes():
            index.save(self)

        return new_val

    def get_id(self):
        """
        Return the primary key for the model instance. If the
        model is unsaved, then this value will be ``None``.
        """
        try:
            return getattr(self, self._primary_key)
        except KeyError:
            return None

    def get_hash_id(self):
        return self._query.get_primary_hash_key(self.get_id())

    def _get_data_dict(self):
        data = {}
        for name, field in self._fields.items():
            if name in self._data:
                data[name] = field.db_value(self._data[name])
        return data

    def to_hash(self):
        """
        Return a :py:class:`Hash` instance corresponding to the
        raw model data.
        """
        return self.__database__.Hash(self.get_hash_id())

    @classmethod
    def create(cls, **kwargs):
        """
        Create a new model instance and save it to the database.
        Values are passed in as keyword arguments.

        Example::

            user = User.create(first_name='Charlie', last_name='Leifer')
        """
        instance = cls(**kwargs)
        instance.save()
        return instance

    @classmethod
    def all(cls):
        """
        Return an iterator that successively yields saved model
        instances. Models are saved in an unordered :py:class:`Set`,
        so the iterator will return them in arbitrary order.

        Example::

            for note in Note.all():
                print note.content

        To return models in sorted order, see :py:meth:`Model.query`.
        Example returning all records, sorted newest to oldest::

            for note in Note.query(order_by=Note.timestamp.desc()):
                print note.timestamp, note.content
        """
        for result in cls._query.all_index():
            yield cls.load(result, convert_key=False)

    @classmethod
    def query(cls, expression=None, order_by=None):
        """
        Return model instances matching the given expression (if
        specified). Additionally, matching instances can be returned
        sorted by field value.

        Example::

            # Get administrators sorted by username.
            admin_users = User.query(
                (User.admin == True),
                order_by=User.username)

            # List blog entries newest to oldest.
            entries = Entry.query(order_by=Entry.timestamp.desc())

            # Perform a complex filter.
            values = StatData.query(
                (StatData.timestamp < datetime.date.today()) &
                ((StatData.type == 'pv') | (StatData.type == 'cv')))

        :param expression: A boolean expression to filter by.
        :param order_by: A field whose value should be used to
            sort returned instances.
        """
        if expression is not None:
            executor = Executor(cls.__database__)
            result = executor.execute(expression)
        else:
            result = cls._query.all_index()

        if order_by is not None:
            desc = False
            if isinstance(order_by, Desc):
                desc = True
                order_by = order_by.node

            alpha = not isinstance(order_by, _ScalarField)
            result = cls.__database__.sort(
                result.key,
                by='*->%s' % order_by.name,
                alpha=alpha,
                desc=desc)
        elif isinstance(result, ZSet):
            result = result.iterator(reverse=True)

        for hash_id in result:
            yield cls.load(hash_id, convert_key=False)

    @classmethod
    def query_delete(cls, expression=None):
        """
        Delete model instances matching the given expression (if
        specified). If no expression is provided, then all model instances
        will be deleted.

        :param expression: A boolean expression to filter by.
        """
        if expression is not None:
            executor = Executor(cls.__database__)
            result = executor.execute(expression)
        else:
            result = cls._query.all_index()

        for hash_id in result:
            cls.load(hash_id, convert_key=False).delete()

    @classmethod
    def get(cls, expression):
        """
        Retrieve the model instance matching the given expression.
        If the number of matching results is not equal to one, then
        a ``ValueError`` will be raised.

        :param expression: A boolean expression to filter by.
        :returns: The matching :py:class:`Model` instance.
        :raises: ``ValueError`` if result set size is not 1.
        """
        executor = Executor(cls.__database__)
        result = executor.execute(expression)
        if len(result) != 1:
            raise ValueError('Got %s results, expected 1.' % len(result))
        return cls.load(result._first_or_any(), convert_key=False)

    @classmethod
    def load(cls, primary_key, convert_key=True):
        """
        Retrieve a model instance by primary key.

        :param primary_key: The primary key of the model instance.
        :returns: Corresponding :py:class:`Model` instance.
        :raises: ``KeyError`` if object with given primary key does
            not exist.
        """
        if convert_key:
            primary_key = cls._query.get_primary_hash_key(primary_key)
        if not cls.__database__.hash_exists(primary_key):
            raise KeyError('Object not found.')
        raw_data = cls.__database__.hgetall(primary_key)
        data = {}
        for name, field in cls._fields.items():
            if isinstance(field, _ContainerField):
                continue
            elif name in raw_data:
                data[name] = field.python_value(raw_data[name])
            elif PY3 and encode(name) in raw_data:
                data[name] = field.python_value(raw_data[encode(name)])
            else:
                data[name] = None

        return cls(**data)

    @classmethod
    def count(cls):
        """
        Return the number of objects in the given collection.
        """
        return len(cls._query.all_index())

    def delete(self, for_update=False):
        """
        Delete the given model instance.
        """
        hash_key = self.get_hash_id()
        try:
            original_instance = self.load(hash_key, convert_key=False)
        except KeyError:
            return

        # Remove from the `all` index.
        all_index = self._query.all_index()
        all_index.remove(hash_key)

        # Remove from the secondary indexes.
        for field in self._indexes:
            for index in field.get_indexes():
                index.remove(original_instance)

        if not for_update:
            for field in self._fields.values():
                if isinstance(field, _ContainerField):
                    field._delete(self)

        # Remove the object itself.
        self.__database__.delete(hash_key)

    def save(self):
        """
        Save the given model instance. If the model does not have
        a primary key value, Walrus will call the primary key field's
        ``generate_key()`` method to attempt to generate a suitable
        value.
        """
        pk_field = self._fields[self._primary_key]
        if not self._data.get(self._primary_key):
            setattr(self, self._primary_key, pk_field._generate_key())
            require_delete = False
        else:
            require_delete = True

        if require_delete:
            self.delete(for_update=True)

        data = self._get_data_dict()
        hash_obj = self.to_hash()
        hash_obj.clear()
        hash_obj.update(data)

        all_index = self._query.all_index()
        all_index.add(self.get_hash_id())

        for field in self._indexes:
            for index in field.get_indexes():
                index.save(self)



import datetime
import random
import sys
import threading
import time
import unittest

from walrus import *
from walrus.query import OP_AND
from walrus.query import OP_OR
try:
    from walrus.tusks.ledisdb import TestWalrusLedis
except ImportError:
    pass
try:
    from walrus.tusks.rlite import TestWalrusLite
except ImportError:
    pass
try:
    from walrus.tusks.vedisdb import TestWalrusVedis
except ImportError:
    pass
from walrus.utils import encode


if sys.version_info[0] == 3:
    unicode = str


db = Database(db=15)


class BaseModel(Model):
    __database__ = db
    __namespace__ = 'test'

class User(BaseModel):
    username = TextField(primary_key=True)

class Note(BaseModel):
    user = TextField(index=True)
    text = TextField()
    timestamp = DateTimeField(default=datetime.datetime.now, index=True)
    tags = JSONField()

class Message(BaseModel):
    content = TextField(fts=True)
    status = IntegerField(default=1, index=True)

class FTSOptions(BaseModel):
    content = TextField(fts=True, stemmer=True)
    metaphone = TextField(fts=True, stemmer=True, metaphone=True)

class Stat(BaseModel):
    key = AutoIncrementField()
    stat_type = ByteField(index=True)
    value = IntegerField(index=True)

class DefaultOption(BaseModel):
    default_empty = JSONField()
    txt = TextField(default='')
    num = IntegerField(default=0)

cache = db.cache(name='test.cache')


@cache.cached(timeout=60)
def now(seed=None):
    return datetime.datetime.now()

class Clock(object):
    @cache.cached_property()
    def now(self):
        return datetime.datetime.now()


class WalrusTestCase(unittest.TestCase):
    def setUp(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def tearDown(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def assertBytes(self, value, expected):
        if isinstance(expected, unicode) and isinstance(value, bytes):
            expected = expected.encode('utf-8')
        return self.assertEqual(value, expected)

    def assertList(self, values, expected):
        for value, item in zip(values, expected):
            self.assertEquivalent(value, item)

        assert len(values) == len(expected)

    def assertDict(self, values, expected):
        for vkey, ekey in zip(sorted(values), sorted(expected)):
            value = values[vkey]
            item = expected[ekey]
            self.assertEquivalent(vkey, ekey)
            self.assertEquivalent(value, item)

    def assertSet(self, values, expected):
        self.assertList(sorted(values), sorted(expected))

    def assertEquivalent(self, lhs, rhs):
        if isinstance(rhs, dict):
            self.assertDict(lhs, rhs)
        elif isinstance(rhs, (list, tuple)):
            self.assertList(lhs, rhs)
        elif isinstance(rhs, set):
            self.assertSet(lhs, rhs)
        else:
            self.assertBytes(lhs, rhs)


class TestModels(WalrusTestCase):
    def create_objects(self):
        for i in range(3):
            u = User.create(username='u%s' % (i + 1))
            for j in range(3):
                Note.create(
                    user=u.username,
                    text='n%s-%s' % (i + 1, j + 1),
                    tags=['t%s' % (k + 1) for k in range(j)])

    def test_store_none(self):
        class Simple(BaseModel):
            text = TextField()
            number = IntegerField()
            normalized = FloatField()

        s = Simple.create(text=None, number=None, normalized=None)
        s_db = Simple.load(s._id)
        self.assertEqual(s_db.text, '')
        self.assertEqual(s_db.number, 0)
        self.assertEqual(s_db.normalized, 0.)

    def test_create(self):
        self.create_objects()
        self.assertEquivalent(
            sorted(user.username for user in User.all()),
            ['u1', 'u2', 'u3'])

        notes = Note.query(Note.user == 'u1')
        self.assertEquivalent(
            sorted(note.text for note in notes),
            ['n1-1', 'n1-2', 'n1-3'])

        notes = sorted(
            Note.query(Note.user == 'u2'),
            key = lambda note: note._id)
        note = notes[2]
        self.assertEquivalent(note.tags, ['t1', 't2'])

    def test_exceptions(self):
        self.assertRaises(KeyError, User.load, 'charlie')
        User.create(username='charlie')
        user = User.load('charlie')
        self.assertEquivalent(user.username, 'charlie')

    def test_query(self):
        self.create_objects()
        notes = Note.query(Note.user == 'u2')
        self.assertEquivalent(
            sorted(note.text for note in notes),
            ['n2-1', 'n2-2', 'n2-3'])

        user = User.get(User.username == 'u3')
        self.assertEquivalent(user._data, {'username': 'u3'})

        self.assertRaises(ValueError, User.get, User.username == 'ux')

    def test_query_with_update(self):
        stat = Stat.create(stat_type='s1', value=1)
        vq = list(Stat.query(Stat.value == 1))
        self.assertEqual(len(vq), 1)
        stat_db = vq[0]
        self.assertEquivalent(stat_db.stat_type, 's1')
        self.assertEqual(stat_db.value, 1)

        stat.value = 2
        stat.save()

        def assertCount(expr, count):
            self.assertEqual(len(list(Stat.query(expr))), count)

        assertCount(Stat.value == 1, 0)
        assertCount(Stat.value == 2, 1)
        assertCount(Stat.stat_type == 's1', 1)

        stat.stat_type = 's2'
        stat.save()

        assertCount(Stat.value == 1, 0)
        assertCount(Stat.value == 2, 1)
        assertCount(Stat.stat_type == 's1', 0)
        assertCount(Stat.stat_type == 's2', 1)

    def test_sorting(self):
        self.create_objects()
        all_notes = [
            'n1-1', 'n1-2', 'n1-3', 'n2-1', 'n2-2', 'n2-3', 'n3-1', 'n3-2',
            'n3-3']

        notes = Note.query(order_by=Note.text)
        self.assertEquivalent([note.text for note in notes], all_notes)

        notes = Note.query(order_by=Note.text.desc())
        self.assertEquivalent(
            [note.text for note in notes],
            all_notes[::-1])

        notes = Note.query(Note.user == 'u2', Note.text)
        self.assertEquivalent(
            [note.text for note in notes],
            ['n2-1', 'n2-2', 'n2-3'])

        notes = Note.query(Note.user == 'u2', Note.text.desc())
        self.assertEquivalent(
            [note.text for note in notes],
            ['n2-3', 'n2-2', 'n2-1'])

    def test_complex_query(self):
        usernames = ['charlie', 'huey', 'mickey', 'zaizee']
        for username in usernames:
            User.create(username=username)

        def assertUsers(expr, expected):
            users = User.query(expr)
            self.assertEquivalent(
                sorted(user.username for user in users),
                sorted(expected))

        assertUsers(User.username == 'charlie', ['charlie'])
        assertUsers(User.username != 'huey', ['charlie', 'mickey', 'zaizee'])
        assertUsers(
            ((User.username == 'charlie') | (User.username == 'mickey')),
            ['charlie', 'mickey'])
        assertUsers(
            (User.username == 'charlie') | (User.username != 'mickey'),
            ['charlie', 'huey', 'zaizee'])
        expr = (
            ((User.username != 'huey') & (User.username != 'zaizee')) |
            (User.username == 'charlie'))
        assertUsers(expr, ['charlie', 'mickey'])

    def test_scalar_query(self):
        data = [
            ('t1', 1),
            ('t1', 2),
            ('t1', 3),
            ('t2', 10),
            ('t2', 11),
            ('t2', 12),
            ('t3', 0),
        ]
        for stat_type, value in data:
            Stat.create(stat_type=stat_type, value=value)

        stat_objects = sorted(
            (stat for stat in Stat.all()),
            key=lambda stat: stat.key)
        self.assertEquivalent([stat._data for stat in stat_objects], [
            {'key': 1, 'stat_type': 't1', 'value': 1},
            {'key': 2, 'stat_type': 't1', 'value': 2},
            {'key': 3, 'stat_type': 't1', 'value': 3},
            {'key': 4, 'stat_type': 't2', 'value': 10},
            {'key': 5, 'stat_type': 't2', 'value': 11},
            {'key': 6, 'stat_type': 't2', 'value': 12},
            {'key': 7, 'stat_type': 't3', 'value': 0},
        ])

        def assertStats(expr, expected):
            stats = Stat.query(expr)
            self.assertEquivalent(
                sorted(stat.key for stat in stats),
                sorted(expected))

        assertStats(Stat.value <= 3, [1, 2, 3, 7])
        assertStats(Stat.value >= 10, [4, 5, 6])
        assertStats(Stat.value < 3, [1, 2, 7])
        assertStats(Stat.value > 10, [5, 6])

        assertStats(Stat.value == 3, [3])
        assertStats(Stat.value >= 13, [])
        assertStats(
            (Stat.value <= 2) | (Stat.key >= 7),
            [1, 2, 7])
        assertStats(
            ((Stat.value <= 2) & (Stat.key >= 7)) | (Stat.value >= 11),
            [5, 6, 7])
        assertStats(
            ((Stat.value <= 2) | (Stat.key >= 7)) & (Stat.stat_type == 't1'),
            [1, 2])

        assertStats(Stat.value.between(2, 11), [2, 3, 4, 5])
        assertStats(Stat.value.between(4, 12), [4, 5, 6])

    def test_full_text_search(self):
        phrases = [
            ('A faith is a necessity to a man. Woe to him who believes in '
             'nothing.'),
            ('All who call on God in true faith, earnestly from the heart, '
             'will certainly be heard, and will receive what they have asked '
             'and desired.'),
            ('Be faithful in small things because it is in them that your '
             'strength lies.'),
            ('Faith consists in believing when it is beyond the power of '
             'reason to believe.'),
            ('Faith has to do with things that are not seen and hope with '
             'things that are not at hand.')]

        for idx, message in enumerate(phrases):
            Message.create(content=message, status=1 + (idx % 2))

        def assertMatches(query, indexes):
            results = [message.content for message in query]
            self.assertEquivalent(results, [phrases[i] for i in indexes])

        def assertSearch(search, indexes):
            assertMatches(
                Message.query(Message.content.match(search)),
                indexes)

        assertSearch('faith', [4, 3, 2, 0, 1])
        assertSearch('faith man', [0])
        assertSearch('things', [4, 2])
        assertSearch('blah', [])

        query = Message.query(
            Message.content.match('faith') & (Message.status == 1))
        assertMatches(query, [4, 2, 0])

    def test_full_text_combined(self):
        phrases = [
            'little bunny foo foo',  # 0, s=1
            'a little green owl',  # 1, s=2
            'the owl was named foo',  # 2, s=1
            'he had a nicotine patch on his wing',  # 3, s=2
            'he was trying to quit smoking',  # 4, s=1
            'the owl was little and green and sweet',  # 5, s=2
            'he dropped presents on my porch',  # 6, s=1
        ]
        index_to_phrase = {}
        for idx, message in enumerate(phrases):
            msg = Message.create(content=message, status=1 + (idx % 2))
            index_to_phrase[idx] = message

        def assertSearch(search, indexes):
            self.assertEquivalent(
                sorted(message.content for message in query),
                sorted(index_to_phrase[idx] for idx in indexes))

        query = Message.query(Message.content.match('little owl'))
        assertSearch(query, [1, 5, 2])  # "little" is ignored (stop word).

        query = Message.query(
            Message.content.match('little owl') &
            Message.content.match('foo'))
        assertSearch(query, [2])

        query = Message.query(
            (Message.content.match('owl') & (Message.status == 1)) |
            (Message.content.match('foo') & (Message.status == 2)))
        assertSearch(query, [2])

        query = Message.query(
            (Message.content.match('green') & (Message.status == 2)) |
            (Message.status == 1))
        assertSearch(query, [0, 2, 4, 6, 1, 5])

        query = Message.query(
            ((Message.status == 2) & Message.content.match('green')) |
            (Message.status == 1))
        assertSearch(query, [0, 2, 4, 6, 1, 5])

    def test_full_text_options(self):
        phrases = [
            'building web applications with python and flask',
            'modern web development with python',
            'unit testing with python',
            'writing better tests for your application',
            'applications for the web',
        ]

        for phrase in phrases:
            FTSOptions.create(content=phrase, metaphone=phrase)

        def assertMatches(search, indexes, use_metaphone=False):
            if use_metaphone:
                field = FTSOptions.metaphone
            else:
                field = FTSOptions.content
            query = FTSOptions.query(field.match(search))
            results = [message.content for message in query]
            self.assertEquivalent(results, [phrases[i] for i in indexes])

        assertMatches('web application', [4, 0])
        assertMatches('web application', [4, 0], True)

        assertMatches('python', [2, 1, 0])
        assertMatches('python', [2, 1, 0], True)

        assertMatches('testing', [3, 2])
        assertMatches('testing', [3, 2], True)

        # Test behavior of the metaphone algorithm.
        assertMatches('python flasck', [0], True)
        assertMatches('pithon devellepment', [], False)
        assertMatches('pithon devellepment', [1], True)
        assertMatches('younit tessts', [2], True)

    def test_fts_query_parser(self):
        messages = [
            'foo green',
            'bar green',
            'baz blue',
            'nug blue',
            'nize yellow',
            'huey greener',
            'mickey greens',
            'zaizee',
        ]
        for message in messages:
            Message.create(content=message)

        def assertMatches(query, expected, default_conjunction=OP_AND):
            expression = Message.content.search(query, default_conjunction)
            messages = Message.query(expression, order_by=Message.content)
            results = [msg.content for msg in messages]
            self.assertEquivalent(results, expected)

        assertMatches('foo', ['foo green'])
        assertMatches('foo OR baz', ['baz blue', 'foo green'])
        assertMatches('green OR blue', [
            'bar green',
            'baz blue',
            'foo green',
            'mickey greens',
            'nug blue',
        ])
        assertMatches('green AND (bar OR mickey OR nize)', [
            'bar green',
            'mickey greens',
        ])
        assertMatches('zaizee OR (blue AND nug) OR (green AND bar)', [
            'bar green',
            'nug blue',
            'zaizee',
        ])
        assertMatches('(blue AND (baz OR (nug OR huey OR mickey))', [
            'baz blue',
            'nug blue',
        ])
        assertMatches(
            '(blue OR foo) AND (green OR (huey OR (baz AND mickey)))',
            ['foo green'])

        assertMatches('(green AND nug) OR (blue AND bar)', [])
        assertMatches('nuglet', [])
        assertMatches('foobar', [])
        assertMatches('', sorted(messages))

    def test_load(self):
        User.create(username='charlie')
        u = User.load('charlie')
        self.assertEquivalent(u._data, {'username': 'charlie'})

    def test_save_delete(self):
        charlie = User.create(username='charlie')
        huey = User.create(username='huey')
        note = Note.create(user='huey', text='n1')
        note.text = 'n1-edited'
        note.save()

        self.assertEquivalent(
            sorted(user.username for user in User.all()),
            ['charlie', 'huey'])

        notes = Note.all()
        self.assertEquivalent([note.text for note in notes], ['n1-edited'])

        charlie.delete()
        self.assertEquivalent([user.username for user in User.all()], ['huey'])

    def test_delete_indexes(self):
        self.assertEquivalent(set(db.keys()), set())

        Message.create(content='charlie message', status=1)
        Message.create(content='huey message', status=2)

        keys = set(db.keys())
        charlie = Message.load(1)
        charlie.delete()

        huey_keys = set(db.keys())
        diff = keys - huey_keys

        make_key = Message._query.make_key
        self.assertEquivalent(diff, set([
            make_key('_id', 'absolute', 1),
            make_key('content', 'absolute', 'charlie message'),
            make_key('content', 'fts', 'charli'),
            make_key('id', 1),
            make_key('status', 'absolute', 1),
        ]))

        # Ensure we cannot query for Charlie, but that we can query for Huey.
        expressions = [
            (Message.status == 1),
            (Message.status != 2),
            (Message._id == 1),
            (Message._id != 2),
            (Message.content == 'charlie message'),
            (Message.content != 'huey message'),
            (Message.content.match('charlie')),
        ]
        for expression in expressions:
            self.assertRaises(ValueError, Message.get, expression)

        expressions = [
            (Message.status == 2),
            (Message.status > 1),
            (Message._id == 2),
            (Message._id != 1),
            (Message.content == 'huey message'),
            (Message.content != 'charlie'),
            (Message.content.match('huey')),
            (Message.content.match('message')),
        ]
        for expression in expressions:
            obj = Message.get(expression)
            self.assertEquivalent(obj._data, {
                '_id': 2,
                'content': 'huey message',
                'status': 2,
            })

        after_filter_keys = set(db.keys())
        symm_diff = huey_keys ^ after_filter_keys
        self.assertTrue(all(decode(key).startswith('temp')
                            for key in symm_diff))

        huey = Message.load(2)
        huey.delete()

        final_keys = set(key for key in db.keys()
                         if not decode(key).startswith('temp'))
        self.assertEquivalent(final_keys, set([make_key('_id', '_sequence')]))

    def test_get_regression(self):
        Message.create(content='huey', status=1)
        Message.create(content='charlie', status=2)

        def assertMessage(msg, data):
            self.assertEqual(msg._data, data)

        huey = {'_id': 1, 'content': 'huey', 'status': 1}
        charlie = {'_id': 2, 'content': 'charlie', 'status': 2}
        assertMessage(Message.load(1), huey)
        assertMessage(Message.load(2), charlie)

        for i in range(3):
            assertMessage(Message.get(Message._id == 1), huey)
            assertMessage(Message.get(Message._id == 2), charlie)

            assertMessage(Message.get(Message.status == 1), huey)
            assertMessage(Message.get(Message.status == 2), charlie)
            assertMessage(Message.get(Message.status != 1), charlie)
            assertMessage(Message.get(Message.status != 2), huey)

            messages = list(Message.query(Message.status == 1))
            self.assertEqual(len(messages), 1)
            assertMessage(messages[0], huey)
            messages = list(Message.query(Message.status != 1))
            self.assertEqual(len(messages), 1)
            assertMessage(messages[0], charlie)

    def test_index_separator(self):
        class CustomSeparator(BaseModel):
            index_separator = '$'
            name = TextField(primary_key=True)
            data = IntegerField(index=True)

        CustomSeparator.create(name='huey.zai', data=3)
        CustomSeparator.create(name='michael.nuggie', data=5)

        keys = sorted(db.keys())
        self.assertEquivalent(keys, [
            # namespace | model : $-delimited indexed data
            'test|customseparator:all',
            'test|customseparator:data$absolute$3',
            'test|customseparator:data$absolute$5',
            'test|customseparator:data$continuous',
            'test|customseparator:id$huey.zai',
            'test|customseparator:id$michael.nuggie',
            'test|customseparator:name$absolute$huey.zai',
            'test|customseparator:name$absolute$michael.nuggie'])

        huey = CustomSeparator.get(CustomSeparator.data < 5)
        self.assertEquivalent(huey.name, 'huey.zai')

        mickey = CustomSeparator.load('michael.nuggie')
        self.assertEquivalent(mickey.data, 5)

    def test_incr(self):
        for i in range(3):
            Stat.create(stat_type='test', value=i)

        s1 = Stat.get(Stat.value == 1)
        res = s1.incr(Stat.value, 5)
        self.assertEqual(res, 6)
        self.assertEqual(s1.value, 6)

        self.assertRaises(ValueError, Stat.get, Stat.value == 1)
        s6 = Stat.get(Stat.value == 6)
        self.assertEqual(s1.key, s6.key)

    def test_count(self):
        self.assertEqual(User.count(), 0)

        for username in ['charlie', 'leslie', 'connor']:
            User.create(username=username)

        self.assertEqual(User.count(), 3)

    def test_query_delete(self):
        for i in range(5):
            u = User.create(username='u%s' % (i + 1))

        User.query_delete((User.username == 'u1') | (User.username == 'u4'))
        usernames = [user.username for user in User.all()]
        self.assertEqual(sorted(usernames), ['u2', 'u3', 'u5'])

        User.query_delete()
        self.assertEqual([user for user in User.all()], [])

    def test_container_field_persistence(self):
        class HashModel(BaseModel):
            data = HashField()
            name = TextField()

        hm1 = HashModel.create(name='hm1')
        hm1.data.update(k1='v1', k2='v2')

        hm2 = HashModel.create(name='hm2')
        hm2.data.update(k3='v3', k4='v4')

        hm1.name = 'hm1-e'
        hm1.save()

        hm1_db = HashModel.load(hm1._id)
        self.assertEquivalent(hm1_db.name, 'hm1-e')
        self.assertEquivalent(hm1.data.as_dict(), {'k1': 'v1', 'k2': 'v2'})

    def test_delete_container_fields(self):
        class HashModel(BaseModel):
            data = HashField()
            name = TextField()

        hm1 = HashModel.create(name='hm1')
        hm1.data.update(k1='v1', k2='v2')

        hm2 = HashModel.create(name='hm2')
        hm2.data.update(k3='v3', k4='v4')

        hm1.delete()
        self.assertEquivalent(hm1.data.as_dict(), {})
        self.assertEquivalent(hm2.data.as_dict(), {'k3': 'v3', 'k4': 'v4'})

    def test_default_is_an_empty_dict(self):
        instance = DefaultOption()
        self.assertTrue(instance.default_empty is None)
        self.assertEqual(instance.num, 0)
        self.assertEqual(instance.txt, '')

    def test_json_storage(self):
        class APIResponse(BaseModel):
            data = JSONField()

        ar = APIResponse(data={'k1': 'v1', 'k2': 'v2'})
        ar.save()

        ar_db = APIResponse.load(ar._id)
        self.assertEqual(ar_db.data, {'k1': 'v1', 'k2': 'v2'})

    def test_pickled_storage(self):
        class PythonData(BaseModel):
            data = PickledField()

        pd = PythonData(data={'k1': ['v1', None, 'v3']})
        pd.save()

        pd_db = PythonData.load(pd._id)
        self.assertEqual(pd_db.data, {'k1': ['v1', None, 'v3']})

        pd2 = PythonData.create(data=None)
        pd2_db = PythonData.load(pd2._id)
        self.assertTrue(pd2_db.data is None)

    def test_boolean_field(self):
        class Account(BaseModel):
            name = TextField(primary_key=True)
            active = BooleanField()
            admin = BooleanField(default=False)

        charlie = Account(name='charlie', active=True, admin=True)
        huey = Account(name='huey', active=False)
        charlie.save()
        huey.save()

        charlie_db = Account.get(Account.name == 'charlie')
        self.assertTrue(charlie_db.active)
        self.assertTrue(charlie_db.admin)

        huey_db = Account.get(Account.name == 'huey')
        self.assertFalse(huey_db.active)
        self.assertFalse(huey_db.admin)

        huey_db.active = True
        huey_db.admin = True
        huey_db.save()

        huey_db2 = Account.get(Account.name == 'huey')
        self.assertTrue(huey_db2.active)
        self.assertTrue(huey_db2.admin)

    def test_uuid(self):
        class Beacon(BaseModel):
            name = TextField(primary_key=True)
            data = UUIDField()

        b1 = Beacon.create(name='alpha', data=uuid.uuid4())
        b2 = Beacon.create(name='bravo', data=uuid.uuid4())
        b3 = Beacon.create(name='charlie')
        b3_db = Beacon.load('charlie')
        b2_db = Beacon.load('bravo')
        b1_db = Beacon.load('alpha')
        self.assertEqual(b1.data, b1_db.data)
        self.assertEqual(b2.data, b2_db.data)
        self.assertTrue(b3.data is None)

    def _test_date_field(self, field_class, dt_func):
        class Event(BaseModel):
            timestamp = field_class(index=True)
            value = TextField()

        events = [
            Event.create(timestamp=dt_func(i), value='e%s' % i)
            for i in range(1, 11)]

        e_db = Event.get(Event._id == events[-1]._id)
        self.assertEqual(e_db.timestamp, dt_func(10))
        self.assertEqual(e_db.value, 'e10')

        events = Event.query(
            (Event.timestamp >= dt_func(3)) &
            (Event.timestamp < dt_func(7)), Event.timestamp)
        ts2value = [(e.timestamp, e.value) for e in events]
        self.assertEqual(ts2value, [
            (dt_func(3), 'e3'),
            (dt_func(4), 'e4'),
            (dt_func(5), 'e5'),
            (dt_func(6), 'e6')])

        e = Event.create(value='ex')
        e_db = Event.load(e._id)
        self.assertTrue(e_db.timestamp is None)
        self.assertEqual(e_db.value, 'ex')

    def test_datetime_field(self):
        dt = lambda day: datetime.datetime(2018, 1, day, 3, 13, 37)
        self._test_date_field(DateTimeField, dt)

    def test_date_field(self):
        dt = lambda day: datetime.date(2018, 1, day)
        self._test_date_field(DateField, dt)


class TestSearchIndex(WalrusTestCase):
    def test_search_index(self):
        phrases = [
            ('A faith is a necessity to a man. Woe to him who believes in '
             'nothing.'),
            ('All who call on God in true faith, earnestly from the heart, '
             'will certainly be heard, and will receive what they have asked '
             'and desired.'),
            ('Be faithful in small things because it is in them that your '
             'strength lies.'),
            ('Faith consists in believing when it is beyond the power of '
             'reason to believe.'),
            ('Faith has to do with things that are not seen and hope with '
             'things that are not at hand.')]

        index = db.Index('test-index')

        for idx, message in enumerate(phrases):
            index.add('doc-%s' % idx, message)

        def assertDocs(query, indexes):
            results = [decode(key['content']) for key in index.search(query)]
            self.assertEqual(results, [phrases[i] for i in indexes])

        assertDocs('faith', [0, 2, 3, 4, 1])
        assertDocs('faith man', [0])
        assertDocs('things', [4, 2])
        assertDocs('blah', [])

    def test_add_remove_update(self):
        data = [
            ('huey cat', {'type': 'cat', 'color': 'white'}),
            ('zaizee cat cat', {'type': 'cat', 'color': 'gray'}),
            ('mickey dog', {'type': 'dog', 'color': 'black'}),
            ('beanie cat', {'type': 'cat', 'color': 'gray'}),
        ]

        idx = db.Index('test-index')
        for i, (content, metadata) in enumerate(data):
            idx.add(str(i), content, **metadata)

        huey, = idx.search('huey')
        self.assertEqual(decode(huey['content']), 'huey cat')
        self.assertEqual(decode(huey['type']), 'cat')
        self.assertEqual(decode(huey['color']), 'white')

        cats = idx.search('cat')
        self.assertEqual([decode(doc['content']) for doc in cats],
                         ['zaizee cat cat', 'huey cat', 'beanie cat'])

        idx.remove('3')  # Poor beanie :(
        zaizee, huey = idx.search('cat')
        self.assertEqual(decode(zaizee['content']), 'zaizee cat cat')
        self.assertEqual(decode(huey['content']), 'huey cat')

        self.assertRaises(KeyError, idx.remove, '3')

        idx.update('1', 'zaizee cat', **{'type': 'kitten'})
        idx.replace('0', 'huey baby cat', **{'type': 'kitten'})

        zaizee, huey = idx.search('cat')
        self.assertEqual(decode(zaizee['content']), 'zaizee cat')
        self.assertEqual(decode(zaizee['type']), 'kitten')
        self.assertEqual(decode(zaizee['color']), 'gray')

        self.assertEqual(decode(huey['content']), 'huey baby cat')
        self.assertEqual(decode(huey['type']), 'kitten')
        self.assertTrue(huey['color'] is None)

    def test_search_phonetic(self):
        data = (
            ('pf', 'python and flask'),
            ('lcp', 'learning cython programming'),
            ('lwd', 'learning web development with flask'),
            ('pwd', 'python web development'))
        data_dict = dict(data)
        idx = db.Index('test-index', metaphone=True)
        for key, content in data:
            idx.add(key, content)

        def assertResults(query, keys):
            result = idx.search(query)
            self.assertEqual([decode(obj['content']) for obj in result],
                             [data_dict[key] for key in keys])

        assertResults('flasck', ['pf', 'lwd'])
        assertResults('pythonn', ['pf', 'pwd'])
        assertResults('sithon', ['lcp'])
        assertResults('webb development', ['pwd', 'lwd'])

        assertResults('sithon OR (flasck AND pythonn)', ['pf', 'lcp'])
        assertResults('garbage', [])

    def test_search_parser(self):
        messages = [
            'foo green',
            'bar green',
            'baz blue',
            'nug blue',
            'nize yellow',
            'huey greener',
            'mickey greens',
            'zaizee',
        ]
        index = db.Index('testing')

        for idx, message in enumerate(messages):
            index.add(str(idx), message)

        def assertMatches(query, expected):
            results = [decode(doc['content']) for doc in index.search(query)]
            self.assertEqual(results, expected)

        assertMatches('foo', ['foo green'])
        assertMatches('foo OR baz', ['foo green', 'baz blue'])
        assertMatches('green OR blue', [
            'foo green',
            'bar green',
            'baz blue',
            'nug blue',
            'mickey greens',
        ])

        assertMatches('green AND (bar OR mickey OR nize)', [
            'bar green',
            'mickey greens',
        ])
        assertMatches('zaizee OR (blue AND nug) OR (green AND bar)', [
            'bar green',
            'nug blue',
            'zaizee',
        ])
        assertMatches('(blue AND (baz OR (nug OR huey OR mickey))', [
            'baz blue',
            'nug blue',
        ])
        assertMatches(
            '(blue OR foo) AND (green OR (huey OR (baz AND mickey)))',
            ['foo green'])

        assertMatches('(green AND nug) OR (blue AND bar)', [])
        assertMatches('nuglet', [])
        assertMatches('foobar', [])
        assertMatches('foo"bar green', ['foo green'])

        results = [decode(doc['content']) for doc in index.search('')]
        self.assertEqual(sorted(results), sorted(messages))


class TestCache(WalrusTestCase):
    def test_cache_apis(self):
        self.assertEqual(cache.get('foo'), None)
        cache.set('foo', 'bar', 60)
        self.assertEqual(cache.get('foo'), 'bar')
        cache.delete('foo')
        self.assertEqual(cache.get('foo'), None)

    def test_cache_decorator(self):
        n1 = now()
        n2 = now(1)
        self.assertNotEqual(n1, n2)
        self.assertEqual(now(), n1)
        self.assertEqual(now(1), n2)

        now.bust(1)
        self.assertNotEqual(now(1), n2)
        self.assertEqual(now(1), now(1))

    def test_cached_property(self):
        c = Clock()
        n1 = c.now
        n2 = c.now
        self.assertEqual(n1, n2)

        del c.now
        n3 = c.now
        self.assertTrue(n1 != n3)
        self.assertEqual(c.now, n3)

    def test_cached_async(self):
        @cache.cache_async()
        def double_value(value):
            return value * 2

        res = double_value(3)
        self.assertEqual(res(), 6)
        self.assertEqual(res(), 6)

        self.assertEqual(double_value(3)(), 6)
        self.assertEqual(double_value(4)(), 8)

    def test_flush_empty_cache(self):
        cache.set('foo', 'bar', 10)
        self.assertEqual(list(cache.keys()), ['test.cache:foo'])
        cache.flush()
        self.assertEqual(list(cache.keys()), [])
        cache.flush()


class TestHash(WalrusTestCase):
    def setUp(self):
        super(TestHash, self).setUp()
        self.hsh = db.Hash('my-hash')

    def test_item_api(self):
        self.hsh['k1'] = 'v1'
        self.assertEquivalent(self.hsh['k1'], 'v1')
        self.assertEquivalent(self.hsh['kx'], None)

        self.hsh['k2'] = 'v2'
        self.hsh['k3'] = 'v3'
        self.assertEquivalent(self.hsh.as_dict(), {
            'k1': 'v1',
            'k2': 'v2',
            'k3': 'v3'})

        del self.hsh['k2']
        self.assertEquivalent(self.hsh.as_dict(), {'k1': 'v1', 'k3': 'v3'})

    def test_dict_apis(self):
        self.hsh.update({'k1': 'v1', 'k2': 'v2'})
        self.hsh.update(k3='v3', k4='v4')
        self.assertEquivalent(sorted(self.hsh.items()), [
            ('k1', 'v1'),
            ('k2', 'v2'),
            ('k3', 'v3'),
            ('k4', 'v4')])
        self.assertEquivalent(sorted(self.hsh.keys()), ['k1', 'k2', 'k3', 'k4'])
        self.assertEquivalent(sorted(self.hsh.values()), ['v1', 'v2', 'v3', 'v4'])

        self.assertEqual(len(self.hsh), 4)
        self.assertTrue('k1' in self.hsh)
        self.assertFalse('kx' in self.hsh)

    def test_search_iter(self):
        self.hsh.update(foo='v1', bar='v2', baz='v3')
        self.assertEquivalent(sorted(self.hsh), [
            ('bar', 'v2'),
            ('baz', 'v3'),
            ('foo', 'v1'),
        ])
        self.assertEquivalent(sorted(self.hsh.search('b*')), [
            ('bar', 'v2'),
            ('baz', 'v3'),
        ])


class TestSet(WalrusTestCase):
    def setUp(self):
        super(TestSet, self).setUp()
        self.set = db.Set('my-set')

    def test_basic_apis(self):
        self.set.add('i1', 'i2', 'i3', 'i2', 'i1')
        self.assertSet(self.set, ['i1', 'i2', 'i3'])

        self.set.remove('i2')
        self.assertSet(self.set, ['i1', 'i3'])

        self.set.remove('ix')
        self.assertSet(self.set, ['i1', 'i3'])

        # Test __contains__
        self.assertTrue('i1' in self.set)
        self.assertFalse('ix' in self.set)

        # Test __iter__.
        self.assertEquivalent(sorted(self.set), ['i1', 'i3'])

        del self.set['i3']
        self.assertSet(self.set, ['i1'])

    def test_combining(self):
        self.set2 = db.Set('my-set2')
        self.set.add(1, 2, 3, 4)
        self.set2.add(3, 4, 5, 6)

        self.assertEquivalent(
            self.set | self.set2,
            set(['1', '2', '3', '4', '5', '6']))

        self.assertEquivalent(
            self.set & self.set2,
            set(['3', '4']))

        self.assertEquivalent(
            self.set - self.set2,
            set(['1', '2']))
        self.assertEquivalent(
            self.set2 - self.set,
            set(['5', '6']))

    def test_combine_store(self):
        self.set2 = db.Set('my-set2')
        self.set.add(1, 2, 3, 4)
        self.set2.add(3, 4, 5, 6)

        s3 = self.set.unionstore('my-set3', self.set2)
        self.assertSet(s3.members(), set(['1', '2', '3', '4', '5', '6']))

        s3 = self.set.interstore('my-set3', self.set2)
        self.assertSet(s3.members(), set(['3', '4']))

        s3 = self.set.diffstore('my-set3', self.set2)
        self.assertSet(s3.members(), set(['1', '2']))

        self.set |= self.set2
        self.assertSet(self.set, ['1', '2', '3', '4', '5', '6'])

        s4 = db.Set('my-set4')
        s4.add('1', '3')
        s3 &= s4
        self.assertSet(s3.members(), set(['1']))

    def test_search(self):
        self.set.add('foo', 'bar', 'baz', 'nug')
        self.assertEquivalent(sorted(self.set.search('b*')), ['bar', 'baz'])

    def test_sort(self):
        values = ['charlie', 'zaizee', 'mickey', 'huey']
        expected = sorted(values)
        self.set.add(*values)
        self.assertEquivalent(self.set.sort(), expected)

        self.set.sort(ordering='DESC', limit=3, store='s_dest')
        res = db.List('s_dest')
        self.assertEquivalent(list(res), ['zaizee', 'mickey', 'huey'])


class TestZSet(WalrusTestCase):
    def setUp(self):
        super(TestZSet, self).setUp()
        self.zs = db.ZSet('my-zset')

    def assertZSet(self, expected):
        self.assertEquivalent(list(self.zs), expected)

    def test_basic_apis(self):
        self.zs.add('i1', 1, 'i2', 2)
        self.assertZSet([('i1', 1), ('i2', 2)])

        self.zs.add('i0', 0)
        self.zs.add('i3', 3)
        self.assertZSet([('i0', 0), ('i1', 1), ('i2', 2), ('i3', 3)])

        self.zs.remove('i1')
        self.zs.remove_by_score(3)
        self.zs.add('i2', -2)
        self.zs.add('i9', 9)
        self.assertZSet([('i2', -2.), ('i0', 0.), ('i9', 9.)])

        # __len__
        self.assertEqual(len(self.zs), 3)

        # __contains__
        self.assertTrue('i0' in self.zs)
        self.assertFalse('i1' in self.zs)

        self.assertEqual(self.zs.score('i2'), -2)
        self.assertEqual(self.zs.score('ix'), None)

        self.assertEqual(self.zs.rank('i0'), 1)
        self.assertEqual(self.zs.rank('i1'), None)

        self.assertEqual(self.zs.count(0, 10), 2)
        self.assertEqual(self.zs.count(-3, 11), 3)

        self.zs.incr('i2')
        self.zs.incr('i0', -2)
        self.assertZSet([('i0', -2.), ('i2', -1.), ('i9', 9.)])

        self.assertEquivalent(self.zs.range_by_score(0, 9), ['i9'])
        self.assertEquivalent(self.zs.range_by_score(-3, 0), ['i0', 'i2'])

    def test_item_apis(self):
        self.zs['i1'] = 1
        self.zs['i0'] = 0
        self.zs['i3'] = 3
        self.zs['i2'] = 2

        self.assertEquivalent(self.zs[0, False], ['i0'])
        self.assertEquivalent(self.zs[0, True], [('i0', 0)])
        self.assertEquivalent(self.zs[2, False], ['i2'])
        self.assertEquivalent(self.zs[2, True], [('i2', 2)])
        self.assertEquivalent(self.zs[-1, True], [('i3', 3)])
        self.assertEquivalent(self.zs[9, True], [])

        self.assertEquivalent(self.zs[0], ['i0'])
        self.assertEquivalent(self.zs[2], ['i2'])
        self.assertEquivalent(self.zs[9], [])

        del self.zs['i1']
        del self.zs['i3']
        self.zs['i2'] = -2
        self.zs['i9'] = 9
        self.assertZSet([('i2', -2.), ('i0', 0.), ('i9', 9.)])

    def test_slicing(self):
        self.zs.add('i1', 1, 'i2', 2, 'i3', 3, 'i0', 0)
        self.assertEquivalent(self.zs[:1, True], [('i0', 0)])
        self.assertEquivalent(self.zs[1:3, False], ['i1', 'i2'])
        self.assertEquivalent(self.zs[1:-1, True], [('i1', 1), ('i2', 2)])

        self.assertEquivalent(self.zs['i1':, False], ['i1', 'i2', 'i3'])
        self.assertEquivalent(self.zs[:'i2', False], ['i0', 'i1'])
        self.assertEquivalent(
            self.zs['i0':'i3', True],
            [('i0', 0), ('i1', 1), ('i2', 2)])
        self.assertRaises(KeyError, self.zs.__getitem__, (slice('i9'), False))
        self.assertEquivalent(self.zs[99:, False], [])

        del self.zs[:'i2']
        self.assertZSet([('i2', 2.), ('i3', 3.)])
        del self.zs[1:]
        self.assertZSet([('i2', 2.)])

    def test_combine_store(self):
        zs2 = db.ZSet('my-zset2')
        self.zs.add(1, 1, 2, 2, 3, 3)
        zs2.add(3, 3, 4, 4, 5, 5)

        zs3 = self.zs.unionstore('my-zset3', zs2)
        self.assertEquivalent(
            list(zs3),
            [('1', 1.), ('2', 2.), ('4', 4.), ('5', 5.), ('3', 6.)])

        zs3 = self.zs.interstore('my-zset3', zs2)
        self.assertEquivalent(list(zs3), [('3', 6.)])

        self.zs |= zs2
        self.assertZSet([
            ('1', 1.), ('2', 2.), ('4', 4.), ('5', 5.), ('3', 6.)])

        zs3 &= zs2
        self.assertEquivalent(list(zs3), [('3', 9.)])

    def test_search(self):
        self.zs.add('foo', 1, 'bar', 2, 'baz', 1, 'nug', 3)
        self.assertEquivalent(
            list(self.zs.search('b*')),
            [('baz', 1.), ('bar', 2.)])

    def test_sort(self):
        values = ['charlie', 3, 'zaizee', 2, 'mickey', 6, 'huey', 3]
        self.zs.add(*values)
        self.assertEquivalent(
            self.zs.sort(),
            ['charlie', 'huey', 'mickey', 'zaizee'])

        self.zs.sort(ordering='DESC', limit=3, store='z_dest')
        res = db.List('z_dest')
        self.assertEquivalent(list(res), ['zaizee', 'mickey', 'huey'])


class TestList(WalrusTestCase):
    def setUp(self):
        super(TestList, self).setUp()
        self.lst = db.List('my-list')

    def test_basic_apis(self):
        self.lst.append('i1')
        self.lst.extend(['i2', 'i3'])
        self.lst.prepend('ix')
        self.assertEquivalent(self.lst, ['ix', 'i1', 'i2', 'i3'])

        self.lst.insert('iy', 'i2', 'before')
        self.lst.insert('iz', 'i2', 'after')
        self.assertEquivalent(self.lst, ['ix', 'i1', 'iy', 'i2', 'iz', 'i3'])

        self.assertEquivalent(self.lst.pop(), 'i3')
        self.assertEquivalent(self.lst.popleft(), 'ix')
        self.assertEqual(len(self.lst), 4)

    def test_item_apis(self):
        self.lst.append('i0')
        self.assertEquivalent(self.lst[0], 'i0')

        self.lst.extend(['i1', 'i2'])
        del self.lst['i1']
        self.assertEquivalent(self.lst, ['i0', 'i2'])

        self.lst[1] = 'i2x'
        self.assertEquivalent(self.lst, ['i0', 'i2x'])

        del self.lst[0]
        self.assertEquivalent(self.lst, ['i2x'])

        del self.lst[99]
        self.assertEquivalent(self.lst, ['i2x'])

        del self.lst['ixxx']
        self.assertEquivalent(self.lst, ['i2x'])

    def test_slicing(self):
        self.lst.extend(['i1', 'i2', 'i3', 'i4'])
        self.assertEquivalent(self.lst[:1], ['i1'])
        self.assertEquivalent(self.lst[:2], ['i1', 'i2'])
        self.assertEquivalent(self.lst[:-1], ['i1', 'i2', 'i3'])
        self.assertEquivalent(self.lst[1:2], ['i2'])
        self.assertEquivalent(self.lst[1:], ['i2', 'i3', 'i4'])

        l = db.List('l1')
        l.extend(range(10))
        del l[1:-1]
        self.assertEquivalent([int(i) for i in l], [1, 2, 3, 4, 5, 6, 7, 8, 9])

        del l[:3]
        self.assertEquivalent([int(i) for i in l], [1, 2, 3])

    def test_sort(self):
        values = ['charlie', 'zaizee', 'mickey', 'huey']
        expected = sorted(values)
        self.lst.extend(values)
        self.assertEquivalent(self.lst.sort(), expected)

        self.lst.sort(ordering='DESC', limit=3, store='l_dest')
        res = db.List('l_dest')
        self.assertEquivalent(list(res), ['zaizee', 'mickey', 'huey'])


class TestArray(WalrusTestCase):
    def setUp(self):
        super(TestArray, self).setUp()
        self.arr = db.Array('my-arr')

    def test_basic_apis(self):
        self.arr.append('i1')
        self.arr.append('i2')
        self.arr.append('i3')
        self.arr.append('i4')
        self.assertEqual(len(self.arr), 4)

        # Indexing works. Invalid indices return None.
        self.assertEquivalent(self.arr[0], 'i1')
        self.assertEquivalent(self.arr[3], 'i4')
        self.assertEqual(self.arr[4], None)

        # Negative indexing works and includes bounds-checking.
        self.assertEquivalent(self.arr[-1], 'i4')
        self.assertEquivalent(self.arr[-4], 'i1')
        self.assertEqual(self.arr[-5], None)

        self.assertEquivalent(self.arr.pop(1), 'i2')
        self.assertEquivalent(list(self.arr), ['i1', 'i3', 'i4'])

        self.assertEquivalent(self.arr.pop(), 'i4')
        self.assertEquivalent(list(self.arr), ['i1', 'i3'])

        self.arr[-1] = 'iy'
        self.arr[0] = 'ix'
        self.assertEquivalent(list(self.arr), ['ix', 'iy'])

        self.assertTrue(encode('iy') in self.arr)
        self.assertFalse(encode('i1') in self.arr)

        self.arr.extend(['foo', 'bar', 'baz'])
        self.assertEquivalent(list(self.arr), ['ix', 'iy', 'foo', 'bar', 'baz'])


class TestWalrus(WalrusTestCase):
    def test_atomic(self):
        def assertDepth(depth):
            self.assertEqual(len(db._transaction_local.pipes), depth)

        assertDepth(0)
        with db.atomic() as p1:
            assertDepth(1)
            with db.atomic() as p2:
                assertDepth(2)
                with db.atomic() as p3:
                    assertDepth(3)
                    p3.pipe.set('k3', 'v3')

                assertDepth(2)
                self.assertEquivalent(db['k3'], 'v3')

                p2.pipe.set('k2', 'v2')

            assertDepth(1)
            self.assertEquivalent(db['k3'], 'v3')
            self.assertEquivalent(db['k2'], 'v2')
            p1.pipe.set('k1', 'v1')

        assertDepth(0)
        self.assertEquivalent(db['k1'], 'v1')
        self.assertEquivalent(db['k2'], 'v2')
        self.assertEquivalent(db['k3'], 'v3')

    def test_atomic_exception(self):
        def do_atomic(k, v, exc=False):
            with db.atomic() as a:
                a.pipe.set(k, v)
                if exc:
                    raise TypeError('foo')

        do_atomic('k', 'v')
        self.assertEquivalent(db['k'], 'v')

        self.assertRaises(TypeError, do_atomic, 'k2', 'v2', True)
        self.assertRaises(KeyError, lambda: db['k2'])
        self.assertEqual(db._transaction_local.pipe, None)

        # Try nested failure.
        with db.atomic() as outer:
            outer.pipe.set('k2', 'v2')
            self.assertRaises(TypeError, do_atomic, 'k3', 'v3', True)

            # Only this will be set.
            outer.pipe.set('k4', 'v4')

        self.assertEquivalent(db._transaction_local.pipe, None)
        self.assertEquivalent(db['k2'], 'v2')
        self.assertRaises(KeyError, lambda: db['k3'])
        self.assertEquivalent(db['k4'], 'v4')

    def test_clear_transaction(self):
        with db.atomic() as a1:
            a1.pipe.set('k1', 'v1')
            with db.atomic() as a2:
                a2.pipe.set('k2', 'v2')
                a2.clear()

        self.assertEquivalent(db['k1'], 'v1')
        self.assertRaises(KeyError, lambda: db['k2'])

        with db.atomic() as a1:
            a1.pipe.set('k3', 'v3')
            with db.atomic() as a2:
                self.assertRaises(KeyError, lambda: db['k3'])

                a2.pipe.set('k4', 'v4')
                a2.clear()

            a1.pipe.set('k5', 'v5')

        self.assertEquivalent(db['k3'], 'v3')
        self.assertRaises(KeyError, lambda: db['k4'])
        self.assertEquivalent(db['k5'], 'v5')

        self.assertEquivalent(db._transaction_local.pipe, None)


class TestCounter(WalrusTestCase):
    def test_counter(self):
        counter_a = db.counter('counter-a')
        counter_b = db.counter('counter-b')

        self.assertEqual(counter_a.value(), 0)
        self.assertEqual(counter_a.incr(), 1)
        self.assertEqual(counter_a.incr(3), 4)
        self.assertEqual(counter_a.value(), 4)

        self.assertEqual(counter_b.value(), 0)
        counter_b += 3
        self.assertEqual(counter_b.value(), 3)
        counter_b = counter_b + counter_a
        self.assertEqual(counter_b.value(), 7)
        counter_b = counter_b - 5
        self.assertEqual(counter_b.value(), 2)


class TestLock(WalrusTestCase):
    def test_lock(self):
        lock_a = db.lock('lock-a')
        lock_b = db.lock('lock-b')
        self.assertTrue(lock_a.acquire())
        self.assertTrue(lock_b.acquire())

        lock_a2 = db.lock('lock-a')
        self.assertFalse(lock_a2.acquire(block=False))
        self.assertFalse(lock_a2.release())
        self.assertNotEqual(lock_a._lock_id, lock_a2._lock_id)

        self.assertFalse(lock_a.acquire(block=False))
        self.assertFalse(lock_b.acquire(block=False))

        t_waiting = threading.Event()
        t_acquired = threading.Event()
        t_acknowledged = threading.Event()

        def wait_for_lock():
            lock_a = db.lock('lock-a')
            t_waiting.set()
            lock_a.acquire()
            t_acquired.set()
            t_acknowledged.wait()
            lock_a.release()

        waiter_t = threading.Thread(target=wait_for_lock)
        waiter_t.start()
        t_waiting.wait()  # Wait until the thread is up and running.

        lock_a.release()
        t_acquired.wait()
        self.assertFalse(lock_a.acquire(block=False))
        t_acknowledged.set()
        waiter_t.join()
        self.assertTrue(lock_a.acquire(block=False))
        lock_a.release()

    def test_lock_ctx_mgr(self):
        lock_a = db.lock('lock-a')
        lock_a2 = db.lock('lock-a')
        with lock_a:
            self.assertFalse(lock_a2.acquire(block=False))
        self.assertTrue(lock_a2.acquire(block=False))

    def test_lock_decorator(self):
        lock = db.lock('lock-a')

        @lock
        def locked():
            lock2 = db.lock('lock-a')
            self.assertFalse(lock2.acquire(block=False))

        locked()

        @lock
        def raise_exception():
            raise ValueError()

        self.assertRaises(ValueError, raise_exception)

        # In the event of an exception, the lock will still be released.
        self.assertTrue(lock.acquire(block=False))


class TestAutocomplete(WalrusTestCase):
    test_data = (
        (1, 'testing python'),
        (2, 'testing python code'),
        (3, 'web testing python code'),
        (4, 'unit tests with python'))

    def setUp(self):
        super(TestAutocomplete, self).setUp()
        self.autocomplete = db.autocomplete()

    def store_test_data(self, id_to_store=None):
        for obj_id, title in self.test_data:
            if id_to_store is None or obj_id == id_to_store:
                self.autocomplete.store(obj_id, title, {
                    'obj_id': obj_id,
                    'title': title,
                    'value': obj_id % 2 == 0 and 'even' or 'odd'})

    def sort_results(self, results):
        return sorted(results, key=lambda item: item['obj_id'])

    def assertResults(self, results, expected):
        self.assertEquivalent(
            [result['obj_id'] for result in results], expected)

    def assertSearch(self, results, expected):
        self.assertEquivalent(list(results), expected)

    def test_search(self):
        self.store_test_data()

        results = self.autocomplete.search('testing python')
        self.assertSearch(results, [
            {'obj_id': 1, 'title': 'testing python', 'value': 'odd'},
            {'obj_id': 2, 'title': 'testing python code', 'value': 'even'},
            {'obj_id': 3, 'title': 'web testing python code', 'value': 'odd'},
        ])

        results = self.autocomplete.search('test')
        self.assertResults(results, [1, 2, 4, 3])

        results = self.autocomplete.search('uni')
        self.assertResults(results, [4])

        self.assertSearch(self.autocomplete.search(''), [])
        self.assertSearch(self.autocomplete.search('missing'), [])
        self.assertSearch(self.autocomplete.search('the'), [])

    def test_boosting(self):
        letters = ('alpha', 'beta', 'gamma', 'delta')
        n = len(letters)
        test_data = []
        for i in range(n * 3):
            obj_id = i + 1
            obj_type = 't%d' % ((i / n) + 1)
            title = 'test %s' % letters[i % n]
            self.autocomplete.store(
                obj_id,
                title,
                {'obj_id': obj_id, 'title': title},
                obj_type)

        def assertBoosts(query, boosts, expected):
            results = self.autocomplete.search(query, boosts=boosts)
            self.assertEqual(
                [result['obj_id'] for result in results],
                expected)

        assertBoosts('alp', None, [1, 5, 9])
        assertBoosts('alp', {'t2': 1.1}, [5, 1, 9])
        assertBoosts('test', {'t3': 1.5, 't2': 1.1}, [
            9, 10, 12, 11, 5, 6, 8, 7, 1, 2, 4, 3])
        assertBoosts('alp', {'t1': 0.5}, [5, 9, 1])
        assertBoosts('alp', {'t1': 1.5, 't3': 1.6}, [9, 1, 5])
        assertBoosts('alp', {'t3': 1.5, '5': 1.6}, [5, 9, 1])

    def test_stored_boosts(self):
        id_to_type = {
            'aaa': 1,
            'aab': 2,
            'aac': 3,
            'aaab': 4,
            'bbbb': 4}
        for obj_id, obj_type in id_to_type.items():
            self.autocomplete.store(obj_id, obj_type=obj_type)

        results = self.autocomplete.search('aa')
        self.assertSearch(results, [
            'aaa',
            'aaab',
            'aab',
            'aac'])

        self.autocomplete.boost_object(obj_type=2, multiplier=2)
        results = self.autocomplete.search('aa')
        self.assertSearch(results, [
            'aab',
            'aaa',
            'aaab',
            'aac'])

        self.autocomplete.boost_object('aac', multiplier=3)
        results = self.autocomplete.search('aa')
        self.assertSearch(results, [
            'aac',
            'aab',
            'aaa',
            'aaab'])

        results = self.autocomplete.search('aa', boosts={'aac': 1.5})
        self.assertSearch(results, [
            'aab',
            'aac',
            'aaa',
            'aaab'])

    def test_limit(self):
        self.store_test_data()
        results = self.autocomplete.search('testing', limit=1)
        self.assertResults(results, [1])

        results = self.autocomplete.search('testing', limit=2)
        self.assertResults(results, [1, 2])

    def test_search_empty(self):
        gen_exp = self.autocomplete.search('')
        self.assertSearch(gen_exp, [])

    def test_chunked(self):
        for i in range(25):
            self.autocomplete.store('foo %s' % (chr(i + ord('a')) * 2))

        ge = self.autocomplete.search('foo', limit=21, chunk_size=5)
        results = list(ge)
        self.assertEqual(len(results), 21)
        self.assertEqual(results[0], 'foo aa')
        self.assertEqual(results[-1], 'foo uu')

    def test_scoring_error(self):
        self.autocomplete.store('aa bb cc')
        self.autocomplete.store('tt cc')

        results = self.autocomplete.search('cc')
        self.assertSearch(results, ['tt cc', 'aa bb cc'])

        self.autocomplete.store('aa b cc')
        results = self.autocomplete.search('cc')
        self.assertSearch(results, ['tt cc', 'aa b cc', 'aa bb cc'])

    def test_simple(self):
        for _, title in self.test_data:
            self.autocomplete.store(title)

        self.assertSearch(self.autocomplete.search('testing'), [
            'testing python',
            'testing python code',
            'web testing python code'])
        self.assertSearch(self.autocomplete.search('code'), [
            'testing python code',
            'web testing python code'])

        self.autocomplete.store('z python code')
        self.assertSearch(self.autocomplete.search('cod'), [
            'testing python code',
            'z python code',
            'web testing python code'])

    def test_sorting(self):
        strings = []
        for i in range(26):
            strings.append('aaaa%s' % chr(i + ord('a')))
            if i > 0:
                strings.append('aaa%sa' % chr(i + ord('a')))

        random.shuffle(strings)
        for s in strings:
            self.autocomplete.store(s)

        results = self.autocomplete.search('aaa')
        self.assertSearch(results, sorted(strings))

        results = self.autocomplete.search('aaa', limit=30)
        self.assertSearch(results, sorted(strings)[:30])

    def test_removing_objects(self):
        self.store_test_data()
        self.autocomplete.remove(1)

        results = self.autocomplete.search('testing')
        self.assertResults(results, [2, 3])

        self.store_test_data(1)
        self.autocomplete.remove(2)

        results = self.autocomplete.search('testing')
        self.assertResults(results, [1, 3])

        # `2` has already been removed
        with self.assertRaises(KeyError):
            self.autocomplete.remove(2)

    def test_tokenize_title(self):
        self.assertEqual(
            self.autocomplete.tokenize_title('abc def ghi'),
            ['abc', 'def', 'ghi'])

        self.assertEqual(self.autocomplete.tokenize_title('a A tHe an a'), [])
        self.assertEqual(self.autocomplete.tokenize_title(''), [])

        self.assertEqual(self.autocomplete.tokenize_title(
            'The Best of times, the blurst of times'),
            ['times', 'blurst', 'times'])

    def test_exists(self):
        self.assertFalse(self.autocomplete.exists('test'))
        self.autocomplete.store('test')
        self.assertTrue(self.autocomplete.exists('test'))

    def test_key_leaks(self):
        initial_key_count = len(db.keys())

        # store the blog "testing python"
        self.store_test_data(1)

        # see how many keys we have in the db - check again in a bit
        key_len = len(db.keys())

        self.store_test_data(2)
        key_len2 = len(db.keys())

        self.assertTrue(key_len != key_len2)
        self.autocomplete.remove(2)

        # back to the original amount of keys
        self.assertEqual(len(db.keys()), key_len)

        self.autocomplete.remove(1)
        self.assertEqual(len(db.keys()), initial_key_count)

    def test_updating(self):
        self.autocomplete.store('id1', 'title baze', 'd1', 't1')
        self.autocomplete.store('id2', 'title nugget', 'd2', 't2')
        self.autocomplete.store('id3', 'title foo', 'd3', 't3')

        results = self.autocomplete.search('tit')
        self.assertSearch(results, ['d1', 'd3', 'd2'])

        # overwrite the data for id1
        self.autocomplete.store('id1', 'title foo', 'D1', 't1')

        results = self.autocomplete.search('tit')
        self.assertSearch(results, ['D1', 'd3', 'd2'])

        # overwrite the data with a new title, will remove the title one refs
        self.autocomplete.store('id1', 'Herple', 'done', 't1')

        results = self.autocomplete.search('tit')
        self.assertSearch(results, ['d3', 'd2'])

        results = self.autocomplete.search('herp')
        self.assertSearch(results, ['done'])

        self.autocomplete.store('id1', 'title baze', 'Done', 't1')
        results = self.autocomplete.search('tit')
        self.assertSearch(results, ['Done', 'd3', 'd2'])

        # this shows that when we don't clean up crap gets left around
        results = self.autocomplete.search('herp')
        self.assertSearch(results, [])

    def test_word_position_ordering(self):
        self.autocomplete.store('aaaa bbbb')
        self.autocomplete.store('bbbb cccc')
        self.autocomplete.store('bbbb aaaa')
        self.autocomplete.store('aaaa bbbb')

        results = self.autocomplete.search('bb')
        self.assertSearch(results, ['bbbb aaaa', 'bbbb cccc', 'aaaa bbbb'])

        results = self.autocomplete.search('aa')
        self.assertSearch(results, ['aaaa bbbb', 'bbbb aaaa'])

        self.autocomplete.store('aabb bbbb')

        results = self.autocomplete.search('bb')
        self.assertSearch(results, [
            'bbbb aaaa',
            'bbbb cccc',
            'aaaa bbbb',
            'aabb bbbb'])

        results = self.autocomplete.search('aa')
        self.assertSearch(results, [
            'aaaa bbbb',
            'aabb bbbb',
            'bbbb aaaa'])

        # Verify issue 9 is fixed.
        self.autocomplete.store('foo one')
        self.autocomplete.store('bar foo one')

        results = self.autocomplete.search('foo')
        self.assertSearch(results, ['foo one', 'bar foo one'])

    def test_return_all_results(self):
        phrases = ('aa bb', 'aa cc', 'bb aa cc', 'bb cc', 'cc aa bb')
        for phrase in phrases:
            self.autocomplete.store(phrase)

        data = self.autocomplete.list_data()
        self.assertEqual(sorted(data), list(phrases))

        titles = self.autocomplete.list_titles()
        self.assertEqual(sorted(data), list(phrases))


class TestRateLimit(WalrusTestCase):
    def setUp(self):
        super(TestRateLimit, self).setUp()
        # Limit to 5 events per second.
        self.rl = db.rate_limit('test-rl', 5, 1)

    def test_rate_limit(self):
        for i in range(5):
            self.assertFalse(self.rl.limit('k1'))

        for i in range(3):
            self.assertTrue(self.rl.limit('k1'))

        self.assertFalse(self.rl.limit('k2'))

    def test_rate_limit_rollover(self):
        rl = db.rate_limit('test-rl2', 3, 100)
        container = db.List('test-rl2:k1')

        now = time.time()
        past = now - 101

        # Simulate two events.
        container.extend([now, now])

        # Third event goes through OK.
        self.assertFalse(rl.limit('k1'))

        # Fourth event is rate-limited.
        self.assertTrue(rl.limit('k1'))

        # There are three timestamps in the container.
        self.assertEqual(len(container), 3)

        # Hand modify the oldest timestamp to appear as if it happened over
        # 100 seconds ago.
        container[-1] = past

        # We can again perform an action.
        self.assertFalse(rl.limit('k1'))

        # We once again have 3 items all within the last 100 seconds, so we
        # are rate-limited.
        self.assertTrue(rl.limit('k1'))

        # There are only 3 items in the container.
        self.assertEqual(len(container), 3)

        # The oldest item is the 2nd we added at the beginning of the test.
        self.assertEqual(float(container[-1]), now)

        # Remove an item and make the 2nd timestamp (oldest) in the past. This
        # gives us 2 actions.
        container.popright()
        container[-1] = past

        self.assertFalse(rl.limit('k1'))
        self.assertFalse(rl.limit('k1'))
        self.assertTrue(rl.limit('k1'))

    def test_decorator(self):
        rl = db.rate_limit('test-rl2', 3, 100)
        container = db.List('test-rl2:fake-key')

        def key_fn(*args, **kwargs):
            return 'fake-key'

        @rl.rate_limited(key_function=key_fn)
        def do_test():
            return 'OK'

        now = time.time()
        container.extend([now, now])

        self.assertEqual(do_test(), 'OK')
        self.assertRaises(RateLimitException, do_test)

        container.popright()
        container[-1] = now - 101

        self.assertEqual(do_test(), 'OK')
        self.assertEqual(do_test(), 'OK')
        self.assertRaises(RateLimitException, do_test)


class TestGraph(WalrusTestCase):
    def setUp(self):
        super(TestGraph, self).setUp()
        # Limit to 5 events per second.
        self.g = db.graph('test-graph')

    def create_graph_data(self):
        data = (
            ('charlie', 'likes', 'huey'),
            ('charlie', 'likes', 'mickey'),
            ('charlie', 'likes', 'zaizee'),
            ('charlie', 'is', 'human'),
            ('connor', 'likes', 'huey'),
            ('connor', 'likes', 'mickey'),
            ('huey', 'eats', 'catfood'),
            ('huey', 'is', 'cat'),
            ('mickey', 'eats', 'anything'),
            ('mickey', 'is', 'dog'),
            ('zaizee', 'eats', 'catfood'),
            ('zaizee', 'is', 'cat'),
        )
        self.g.store_many(data)

    def create_friends(self):
        data = (
            ('charlie', 'friend', 'huey'),
            ('huey', 'friend', 'charlie'),
            ('huey', 'friend', 'mickey'),
            ('zaizee', 'friend', 'charlie'),
            ('zaizee', 'friend', 'mickey'),
            ('mickey', 'friend', 'nuggie'),
        )
        for item in data:
            self.g.store(*item)

    def test_search_extended(self):
        self.create_graph_data()
        X = self.g.v.x
        Y = self.g.v.y
        Z = self.g.v.z
        result = self.g.search(
            (X, 'likes', Y),
            (Y, 'is', 'cat'),
            (Z, 'likes', Y))
        self.assertEqual(result['x'], set(['charlie', 'connor']))
        self.assertEqual(result['y'], set(['huey', 'zaizee']))
        self.assertEqual(result['z'], set(['charlie', 'connor']))

        self.g.store_many((
            ('charlie', 'likes', 'connor'),
            ('connor', 'likes', 'charlie'),
            ('connor', 'is', 'baby'),
            ('connor', 'is', 'human'),
            ('nash', 'is', 'baby'),
            ('nash', 'is', 'human'),
            ('connor', 'lives', 'ks'),
            ('nash', 'lives', 'nv'),
            ('charlie', 'lives', 'ks')))

        result = self.g.search(
            ('charlie', 'likes', X),
            (X, 'is', 'baby'),
            (X, 'lives', 'ks'))
        self.assertEqual(result, {'x': set(['connor'])})

        result = self.g.search(
            (X, 'is', 'baby'),
            (X, 'likes', Y),
            (Y, 'lives', 'ks'))
        self.assertEqual(result, {
            'x': set(['connor']),
            'y': set(['charlie']),
        })

    def assertTriples(self, result, expected):
        result = list(result)
        self.assertEqual(len(result), len(expected))
        for i1, i2 in zip(result, expected):
            self.assertEqual(
                (i1['s'], i1['p'], i1['o']), i2)

    def test_query(self):
        self.create_graph_data()
        res = self.g.query('charlie', 'likes')
        self.assertTriples(res, (
            ('charlie', 'likes', 'huey'),
            ('charlie', 'likes', 'mickey'),
            ('charlie', 'likes', 'zaizee'),
        ))

        res = self.g.query(p='is', o='cat')
        self.assertTriples(res, (
            ('huey', 'is', 'cat'),
            ('zaizee', 'is', 'cat'),
        ))

        res = self.g.query(s='huey')
        self.assertTriples(res, (
            ('huey', 'eats', 'catfood'),
            ('huey', 'is', 'cat'),
        ))

        res = self.g.query(o='huey')
        self.assertTriples(res, (
            ('charlie', 'likes', 'huey'),
            ('connor', 'likes', 'huey'),
        ))

    def test_search(self):
        self.create_graph_data()
        X = self.g.v('x')
        result = self.g.search(
            {'s': 'charlie', 'p': 'likes', 'o': X},
            {'s': X, 'p': 'eats', 'o': 'catfood'},
            {'s': X, 'p': 'is', 'o': 'cat'})
        self.assertEqual(result, {'x': set(['huey', 'zaizee'])})

    def test_search_simple(self):
        self.create_friends()
        X = self.g.v('x')
        result = self.g.search({'s': X, 'p': 'friend', 'o': 'charlie'})
        self.assertEqual(result, {'x': set(['huey', 'zaizee'])})

    def test_search_2var(self):
        self.create_friends()
        X = self.g.v('x')
        Y = self.g.v('y')

        result = self.g.search(
            {'s': X, 'p': 'friend', 'o': 'charlie'},
            {'s': Y, 'p': 'friend', 'o': X})
        self.assertEqual(result, {
            'x': set(['huey']),
            'y': set(['charlie']),
        })

        result = self.g.search(
            ('charlie', 'friend', X),
            (X, 'friend', Y),
            (Y, 'friend', 'nuggie'))
        self.assertEqual(result, {
            'x': set(['huey']),
            'y': set(['mickey']),
        })

        result = self.g.search(
            ('huey', 'friend', X),
            (X, 'friend', Y))
        self.assertEqual(result['y'], set(['huey', 'nuggie']))

    def test_search_mutual(self):
        self.create_friends()
        X = self.g.v('x')
        Y = self.g.v('y')

        result = self.g.search(
            {'s': X, 'p': 'friend', 'o': Y},
            {'s': Y, 'p': 'friend', 'o': X})
        self.assertEqual(result['y'], set(['charlie', 'huey']))


if __name__ == '__main__':
    unittest.main(argv=sys.argv)



import datetime

from walrus import *


class TestHelper(object):
    def test_simple_string_ops(self):
        self.assertTrue(self.db.set('name', 'charlie'))
        self.assertEqual(self.db.get('name'), 'charlie')
        self.assertIsNone(self.db.get('not-exist'))

        self.assertFalse(self.db.setnx('name', 'huey'))
        self.db.setnx('friend', 'zaizee')
        self.assertEqual(self.db['name'], 'charlie')
        self.assertEqual(self.db['friend'], 'zaizee')

        self.assertTrue(self.db.mset({'k1': 'v1', 'k2': 'v2'}))
        res = self.db.mget('k1', 'k2')
        self.assertEqual(res, ['v1', 'v2'])

        self.db.append('k1', 'xx')
        self.assertEqual(self.db['k1'], 'v1xx')

        del self.db['counter']
        self.assertEqual(self.db.incr('counter'), 1)
        self.assertEqual(self.db.incr('counter', 5), 6)
        self.assertEqual(self.db.decr('counter', 2), 4)

        self.assertEqual(self.db.getrange('name', 3, 5), 'rli')
        self.assertEqual(self.db.getset('k2', 'baze'), 'v2')
        self.assertEqual(self.db['k2'], 'baze')
        self.assertEqual(self.db.strlen('name'), 7)

        self.db['data'] = '\x07'
        self.assertEqual(self.db.bitcount('data'), 3)

        del self.db['name']
        self.assertIsNone(self.db.get('name'))
        self.assertRaises(KeyError, lambda: self.db['name'])

        self.assertFalse('name' in self.db)
        self.assertTrue('k1' in self.db)

    def test_simple_hash(self):
        h = self.db.Hash('hash_obj')
        h.clear()

        h['k1'] = 'v1'
        h.update({'k2': 'v2', 'k3': 'v3'})
        self.assertEqual(h.as_dict(), {'k1': 'v1', 'k2': 'v2', 'k3': 'v3'})

        self.assertEqual(h['k2'], 'v2')
        self.assertIsNone(h['k4'])
        self.assertTrue('k2' in h)
        self.assertFalse('k4' in h)

        del h['k2']
        del h['k4']
        self.assertEqual(sorted(h.keys()), ['k1', 'k3'])
        self.assertEqual(sorted(h.values()), ['v1', 'v3'])
        self.assertEqual(len(h), 2)

        self.assertEqual(h['k1', 'k2', 'k3'], ['v1', None, 'v3'])

        self.assertEqual(h.incr('counter'), 1)
        self.assertEqual(h.incr('counter', 3), 4)

    def test_simple_list(self):
        l = self.db.List('list_obj')
        l.clear()

        l.append('charlie')
        l.extend(['mickey', 'huey', 'zaizee'])
        self.assertEqual(l[1], 'mickey')
        self.assertEqual(l[-1], 'zaizee')

        self.assertEqual(l[:], ['charlie', 'mickey', 'huey', 'zaizee'])
        self.assertEqual(l[1:-1], ['mickey', 'huey'])
        self.assertEqual(l[2:], ['huey', 'zaizee'])
        self.assertEqual(len(l), 4)

        l.prepend('nuggie')
        l.popright()
        l.popright()
        self.assertEqual([item for item in l], ['nuggie', 'charlie', 'mickey'])

        self.assertEqual(l.popleft(), 'nuggie')
        self.assertEqual(l.popright(), 'mickey')

        l.clear()
        self.assertEqual(list(l), [])
        self.assertIsNone(l.popleft())

    def test_simple_set(self):
        s = self.db.Set('set_obj')
        s.clear()

        self.assertTrue(s.add('charlie'))
        self.assertFalse(s.add('charlie'))
        s.add('huey', 'mickey')
        self.assertEqual(len(s), 3)
        self.assertTrue('huey' in s)
        self.assertFalse('xx' in s)
        self.assertEqual(s.members(), set(['charlie', 'huey', 'mickey']))

        del s['huey']
        del s['xx']
        self.assertEqual(s.members(), set(['charlie', 'mickey']))

        n1 = self.db.Set('n1')
        n2 = self.db.Set('n2')
        n1.add(*range(5))
        n2.add(*range(3, 7))

        self.assertEqual(n1 - n2, set(['0', '1', '2']))
        self.assertEqual(n2 - n1, set(['5', '6']))
        self.assertEqual(n1 | n2, set(map(str, range(7))))
        self.assertEqual(n1 & n2, set(['3', '4']))

        n1.diffstore('ndiff', n2)
        ndiff = self.db.Set('ndiff')
        self.assertEqual(ndiff.members(), set(['0', '1', '2']))

        n1.interstore('ninter', n2)
        ninter = self.db.Set('ninter')
        self.assertEqual(ninter.members(), set(['3', '4']))

    def test_zset(self):
        zs = self.db.ZSet('zset_obj')
        zs.clear()

        zs.add('charlie', 31, 'huey', 3, 'mickey', 6, 'zaizee', 3, 'nuggie', 0)
        self.assertEqual(zs[1], ['huey'])
        self.assertEqual(zs[1, True], [('huey', 3)])

        self.assertEqual(
            zs[:],
            ['nuggie', 'huey', 'zaizee', 'mickey', 'charlie'])
        self.assertEqual(zs[:2], ['nuggie', 'huey'])
        self.assertEqual(zs[1:3, True], [('huey', 3), ('zaizee', 3)])
        self.assertEqual(zs['huey':'charlie'], ['huey', 'zaizee', 'mickey'])

        self.assertEqual(len(zs), 5)
        self.assertTrue('charlie' in zs)
        self.assertFalse('xx' in zs)

        self.assertEqual(zs.score('charlie'), 31.)
        self.assertIsNone(zs.score('xx'))

        self.assertEqual(zs.rank('mickey'), 3)
        self.assertIsNone(zs.rank('xx'))

        self.assertEqual(zs.count(0, 5), 3)
        self.assertEqual(zs.count(6, 31), 2)
        self.assertEqual(zs.count(6, 30), 1)

        zs.incr('mickey')
        self.assertEqual(zs.score('mickey'), 7.)

        self.assertEqual(zs.range_by_score(0, 5), ['nuggie', 'huey', 'zaizee'])

        zs.remove('nuggie')
        self.assertEqual(zs[:2], ['huey', 'zaizee'])

        del zs['mickey']
        self.assertEqual(zs[:], ['huey', 'zaizee', 'charlie'])
        self.assertEqual(len(zs), 3)

        zs.remove_by_score(2, 4)
        self.assertEqual(zs[:], ['charlie'])

        zs.add('huey', 4, 'zaizee', 3, 'beanie', 8)
        zs.remove_by_rank(2)
        self.assertEqual(zs[:], ['zaizee', 'huey', 'charlie'])

        self.assertRaises(KeyError, lambda: zs['xx':])

        z1 = self.db.ZSet('z1')
        z2 = self.db.ZSet('z2')
        z1.add(1, 1, 2, 2, 3, 3)
        z2.add(3, 3, 4, 4, 5, 5)
        z3 = z1.unionstore('z3', z2)
        self.assertEqual(z3[:], ['1', '2', '4', '5', '3'])

        z3 = z1.interstore('z3', z2)
        self.assertEqual(z3[:], ['3'])

    def test_models(self):
        class User(Model):
            __database__ = self.db
            username = TextField(primary_key=True)
            value = IntegerField(index=True)

        for i, username in enumerate(('charlie', 'huey', 'zaizee', 'mickey')):
            User.create(username=username, value=i)

        charlie = User.load('charlie')
        self.assertEqual(charlie.username, 'charlie')
        self.assertEqual(charlie.value, 0)

        query = User.query(
            (User.username == 'charlie') |
            (User.username == 'huey'))
        users = [user.username for user in query]
        self.assertEqual(sorted(users), ['charlie', 'huey'])

    def test_cache(self):
        cache = self.db.cache(name='test-cache')

        @cache.cached(timeout=10)
        def now(seed=None):
            return datetime.datetime.now()

        dt1 = now()
        self.assertEqual(now(), dt1)
        self.assertNotEqual(now(1), dt1)
        self.assertEqual(now(1), now(1))



"""
Lightweight Python utilities for working with Redis.
"""

__author__ = 'Charles Leifer'
__license__ = 'MIT'
__version__ = '0.5.1'

#               ___
#            .-9 9 `\
#          =(:(::)=  ;
#            ||||     \
#            ||||      `-.
#           ,\|\|         `,
#          /                \
#         ;                  `'---.,
#         |                         `\
#         ;                     /     |
#         \                    |      /
#  jgs     )           \  __,.--\    /
#       .-' \,..._\     \`   .-'  .-'
#      `-=``      `:    |   /-/-/`
#                   `.__/

from walrus.autocomplete import Autocomplete
from walrus.cache import Cache
from walrus.containers import Array
from walrus.containers import Container
from walrus.containers import Hash
from walrus.containers import HyperLogLog
from walrus.containers import List
from walrus.containers import Set
from walrus.containers import ZSet
from walrus.database import Database
from walrus.fts import Index
from walrus.lock import Lock
from walrus.models import *
from walrus.rate_limit import RateLimit
from walrus.rate_limit import RateLimitException

# Friendly alias.
Walrus = Database



class Counter(object):
    """
    Simple counter.
    """
    def __init__(self, database, name):
        """
        :param database: A walrus ``Database`` instance.
        :param str name: The name for the counter.
        """
        self.database = database
        self.name = name
        self.key = 'counter:%s' % self.name
        if self.key not in self.database:
            self.database[self.key] = 0

    def decr(self, decr_by=1):
        return self.database.decr(self.key, decr_by)

    def incr(self, incr_by=1):
        return self.database.incr(self.key, incr_by)

    def value(self):
        return int(self.database[self.key])

    def _op(self, method, other):
        if isinstance(other, Counter):
            other = other.value()
        if not isinstance(other, int):
            raise TypeError('Cannot add %s, not an integer.' % other)
        method(other)
        return self

    def __iadd__(self, other):
        return self._op(self.incr, other)

    def __isub__(self, other):
        return self._op(self.decr, other)

    __add__ = __iadd__
    __sub__ = __isub__



from functools import wraps
import glob
import os
import threading
import uuid

try:
    from redis import Redis
except ImportError:
    Redis = object

from walrus.autocomplete import Autocomplete
from walrus.cache import Cache
from walrus.containers import Array
from walrus.containers import Hash
from walrus.containers import HyperLogLog
from walrus.containers import List
from walrus.containers import Set
from walrus.containers import ZSet
from walrus.counter import Counter
from walrus.fts import Index
from walrus.graph import Graph
from walrus.lock import Lock
from walrus.rate_limit import RateLimit


class TransactionLocal(threading.local):
    def __init__(self, **kwargs):
        super(TransactionLocal, self).__init__(**kwargs)
        self.pipes = []

    @property
    def pipe(self):
        if len(self.pipes):
            return self.pipes[-1]

    def commit(self):
        pipe = self.pipes.pop()
        return pipe.execute()

    def abort(self):
        pipe = self.pipes.pop()
        pipe.reset()


class Database(Redis):
    """
    Redis-py client with some extras.
    """
    def __init__(self, *args, **kwargs):
        """
        :param args: Arbitrary positional arguments to pass to the
            base ``Redis`` instance.
        :param kwargs: Arbitrary keyword arguments to pass to the
            base ``Redis`` instance.
        :param str script_dir: Path to directory containing walrus
            scripts.
        """
        script_dir = kwargs.pop('script_dir', None)
        super(Database, self).__init__(*args, **kwargs)
        self.__mapping = {
            'list': self.List,
            'set': self.Set,
            'zset': self.ZSet,
            'hash': self.Hash}
        self._transaction_local = TransactionLocal()
        self._transaction_lock = threading.RLock()
        self.init_scripts(script_dir=script_dir)

    def get_transaction(self):
        with self._transaction_lock:
            local = self._transaction_local
            local.pipes.append(self.pipeline())
            return local.pipe

    def commit_transaction(self):
        """
        Commit the currently active transaction (Pipeline). If no
        transaction is active in the current thread, an exception
        will be raised.

        :returns: The return value of executing the Pipeline.
        :raises: ``ValueError`` if no transaction is active.
        """
        with self._transaction_lock:
            local = self._transaction_local
            if not local.pipes:
                raise ValueError('No transaction is currently active.')
            return local.commit()

    def clear_transaction(self):
        """
        Clear the currently active transaction (if exists). If the
        transaction stack is not empty, then a new pipeline will
        be initialized.

        :returns: No return value.
        :raises: ``ValueError`` if no transaction is active.
        """
        with self._transaction_lock:
            local = self._transaction_local
            if not local.pipes:
                raise ValueError('No transaction is currently active.')
            local.abort()

    def atomic(self):
        return _Atomic(self)

    def init_scripts(self, script_dir=None):
        self._scripts = {}
        if not script_dir:
            script_dir = os.path.join(os.path.dirname(__file__), 'scripts')
        for filename in glob.glob(os.path.join(script_dir, '*.lua')):
            with open(filename, 'r') as fh:
                script_obj = self.register_script(fh.read())
                script_name = os.path.splitext(os.path.basename(filename))[0]
                self._scripts[script_name] = script_obj

    def run_script(self, script_name, keys=None, args=None):
        """
        Execute a walrus script with the given arguments.

        :param script_name: The base name of the script to execute.
        :param list keys: Keys referenced by the script.
        :param list args: Arguments passed in to the script.
        :returns: Return value of script.

        .. note:: Redis scripts require two parameters, ``keys``
            and ``args``, which are referenced in lua as ``KEYS``
            and ``ARGV``.
        """
        return self._scripts[script_name](keys, args)

    def get_temp_key(self):
        """
        Generate a temporary random key using UUID4.
        """
        return 'temp.%s' % uuid.uuid4()

    def __iter__(self):
        """
        Iterate over the keys of the selected database.
        """
        return iter(self.scan_iter())

    def search(self, pattern):
        """
        Search the keyspace of the selected database using the
        given search pattern.

        :param str pattern: Search pattern using wildcards.
        :returns: Iterator that yields matching keys.
        """
        return self.scan_iter(pattern)

    def get_key(self, key):
        """
        Return a rich object for the given key. For instance, if
        a hash key is requested, then a :py:class:`Hash` will be
        returned.

        :param str key: Key to retrieve.
        :returns: A hash, set, list, zset or array.
        """
        return self.__mapping.get(self.type(key), self.__getitem__)(key)

    def hash_exists(self, key):
        return self.exists(key)

    def autocomplete(self, namespace='autocomplete', **kwargs):
        return Autocomplete(self, namespace, **kwargs)

    def cache(self, name='cache', default_timeout=3600):
        """
        Create a :py:class:`Cache` instance.

        :param str name: The name used to prefix keys used to
            store cached data.
        :param int default_timeout: The default key expiry.
        :returns: A :py:class:`Cache` instance.
        """
        return Cache(self, name=name, default_timeout=default_timeout)

    def counter(self, name):
        """
        Create a :py:class:`Counter` instance.

        :param str name: The name used to store the counter's value.
        :returns: A :py:class:`Counter` instance.
        """
        return Counter(self, name=name)

    def graph(self, name, *args, **kwargs):
        """
        Creates a :py:class:`Graph` instance.

        :param str name: The namespace for the graph metadata.
        :returns: a :py:class:`Graph` instance.
        """
        return Graph(self, name, *args, **kwargs)

    def lock(self, name, ttl=None, lock_id=None):
        """
        Create a named :py:class:`Lock` instance. The lock implements
        an API similar to the standard library's ``threading.Lock``,
        and can also be used as a context manager or decorator.

        :param str name: The name of the lock.
        :param int ttl: The time-to-live for the lock in milliseconds
            (optional). If the ttl is ``None`` then the lock will not
            expire.
        :param str lock_id: Optional identifier for the lock instance.
        """
        return Lock(self, name, ttl, lock_id)

    def rate_limit(self, name, limit=5, per=60, debug=False):
        """
        Rate limit implementation. Allows up to `limit` of events every `per`
        seconds.

        See :ref:`rate-limit` for more information.
        """
        return RateLimit(self, name, limit, per, debug)

    def Index(self, name, **options):
        """
        Create a :py:class:`Index` full-text search index with the given
        name and options.
        """
        return Index(self, name, **options)

    def List(self, key):
        """
        Create a :py:class:`List` instance wrapping the given key.
        """
        return List(self, key)

    def Hash(self, key):
        """
        Create a :py:class:`Hash` instance wrapping the given key.
        """
        return Hash(self, key)

    def Set(self, key):
        """
        Create a :py:class:`Set` instance wrapping the given key.
        """
        return Set(self, key)

    def ZSet(self, key):
        """
        Create a :py:class:`ZSet` instance wrapping the given key.
        """
        return ZSet(self, key)

    def HyperLogLog(self, key):
        """
        Create a :py:class:`HyperLogLog` instance wrapping the given
        key.
        """
        return HyperLogLog(self, key)

    def Array(self, key):
        """
        Create a :py:class:`Array` instance wrapping the given key.
        """
        return Array(self, key)

    def listener(self, channels=None, patterns=None, async=False):
        """
        Decorator for wrapping functions used to listen for Redis
        pub-sub messages.

        The listener will listen until the decorated function
        raises a ``StopIteration`` exception.

        :param list channels: Channels to listen on.
        :param list patterns: Patterns to match.
        :param bool async: Whether to start the listener in a
            separate thread.
        """
        def decorator(fn):
            _channels = channels or []
            _patterns = patterns or []

            @wraps(fn)
            def inner():
                pubsub = self.pubsub()

                def listen():
                    for channel in _channels:
                        pubsub.subscribe(channel)
                    for pattern in _patterns:
                        pubsub.psubscribe(pattern)

                    for data_dict in pubsub.listen():
                        try:
                            ret = fn(**data_dict)
                        except StopIteration:
                            pubsub.close()
                            break

                if async:
                    worker = threading.Thread(target=listen)
                    worker.start()
                    return worker
                else:
                    listen()

            return inner
        return decorator

    def stream_log(self, callback, connection_id='monitor'):
        """
        Stream Redis activity one line at a time to the given
        callback.

        :param callback: A function that accepts a single argument,
            the Redis command.
        """
        conn = self.connection_pool.get_connection(connection_id, None)
        conn.send_command('monitor')
        while callback(conn.read_response()):
            pass


class _Atomic(object):
    def __init__(self, db):
        self.db = db

    @property
    def pipe(self):
        return self.db._transaction_local.pipe

    def __enter__(self):
        self.db.get_transaction()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.clear(False)
        else:
            self.commit(False)

    def commit(self, begin_new=True):
        ret = self.db.commit_transaction()
        if begin_new:
            self.db.get_transaction()
        return ret

    def clear(self, begin_new=True):
        self.db.clear_transaction()
        if begin_new:
            self.db.get_transaction()



import sys
import unittest

from ledis import Ledis
from ledis.client import Token

from walrus import *
from walrus.containers import chainable_method
from walrus.tusks.helpers import TestHelper


class Scannable(object):
    def _scan(self, cmd, match=None, count=None, ordering=None, limit=None):
        parts = [self.key, '']
        if match:
            parts.extend([Token('MATCH'), match])
        if count:
            parts.extend([Token('COUNT'), count])
        if ordering:
            parts.append(Token(ordering.upper()))
        return self._execute_scan(self.database, cmd, parts, limit)

    def _execute_scan(self, database, cmd, parts, limit=None):
        idx = 0
        while True:
            cursor, rows = database.execute_command(cmd, *parts)
            for row in rows:
                idx += 1
                if limit and idx > limit:
                    cursor = None
                    break
                yield row
            if cursor:
                parts[1] = cursor
            else:
                break


class Sortable(object):
    def _sort(self, cmd, pattern=None, limit=None, offset=None,
              get_pattern=None, ordering=None, alpha=True, store=None):
        parts = [self.key]
        def add_kw(kw, param):
            if param is not None:
                parts.extend([Token(kw), param])
        add_kw('BY', pattern)
        if limit or offset:
            offset = offset or 0
            limit = limit or 'Inf'
            parts.extend([Token('LIMIT'), offset, limit])
        add_kw('GET', get_pattern)
        if ordering:
            parts.append(Token(ordering))
        if alpha:
            parts.append(Token('ALPHA'))
        add_kw('STORE', store)
        return self.database.execute_command(cmd, *parts)


class LedisHash(Scannable, Hash):
    @chainable_method
    def clear(self):
        self.database.hclear(self.key)

    @chainable_method
    def expire(self, ttl=None):
        if ttl is not None:
            self.database.hexpire(self.key, ttl)
        else:
            self.database.hpersist(self.key)

    def __iter__(self):
        return self._scan('XHSCAN')

    def scan(self, match=None, count=None, ordering=None, limit=None):
        if limit is not None:
            limit *= 2  # Hashes yield 2 values.
        return self._scan('XHSCAN', match, count, ordering, limit)


class LedisList(Sortable, List):
    @chainable_method
    def clear(self):
        self.database.lclear(self.key)

    def __setitem__(self, idx, value):
        raise TypeError('Ledis does not support setting values by index.')

    @chainable_method
    def expire(self, ttl=None):
        if ttl is not None:
            self.database.lexpire(self.key, ttl)
        else:
            self.database.lpersist(self.key)

    def sort(self, *args, **kwargs):
        return self._sort('XLSORT', *args, **kwargs)


class LedisSet(Scannable, Sortable, Set):
    @chainable_method
    def clear(self):
        self.database.sclear(self.key)

    @chainable_method
    def expire(self, ttl=None):
        if ttl is not None:
            self.database.sexpire(self.key, ttl)
        else:
            self.database.spersist(self.key)

    def __iter__(self):
        return self._scan('XSSCAN')

    def scan(self, match=None, count=None, ordering=None, limit=None):
        return self._scan('XSSCAN', match, count, ordering, limit)

    def sort(self, *args, **kwargs):
        return self._sort('XSSORT', *args, **kwargs)


class LedisZSet(Scannable, Sortable, ZSet):
    @chainable_method
    def clear(self):
        self.database.zclear(self.key)

    @chainable_method
    def expire(self, ttl=None):
        if ttl is not None:
            self.database.zexpire(self.key, ttl)
        else:
            self.database.zpersist(self.key)

    def __iter__(self):
        return self._scan('XZSCAN')

    def scan(self, match=None, count=None, ordering=None, limit=None):
        if limit:
            limit *= 2
        return self._scan('XZSCAN', match, count, ordering, limit)

    def sort(self, *args, **kwargs):
        return self._sort('XZSORT', *args, **kwargs)


class LedisBitSet(Container):
    def clear(self):
        self.database.delete(self.key)

    def __getitem__(self, idx):
        return self.database.execute_command('GETBIT', self.key, idx)

    def __setitem__(self, idx, value):
        return self.database.execute_command('SETBIT', self.key, idx, value)

    def pos(self, bit, start=None, end=None):
        pieces = ['BITPOS', self.key, bit]
        if start or end:
            pieces.append(start or 0)
        if end:
            pieces.append(end)
        return self.database.execute_command(*pieces)

    def __iand__(self, other):
        self.database.execute_command(
            'BITOP',
            'AND',
            self.key,
            self.key,
            other.key)
        return self

    def __ior__(self, other):
        self.database.execute_command(
            'BITOP',
            'OR',
            self.key,
            self.key,
            other.key)
        return self

    def __ixor__(self, other):
        self.database.execute_command(
            'BITOP',
            'XOR',
            self.key,
            self.key,
            other.key)
        return self

    def __str__(self):
        return self.database[self.key]

    __unicode__ = __str__


class WalrusLedis(Ledis, Scannable, Walrus):
    def __init__(self, *args, **kwargs):
        super(WalrusLedis, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        self.set(key, value)

    def setex(self, name, value, time):
        return super(WalrusLedis, self).setex(name, time, value)

    def zadd(self, key, *args, **kwargs):
        if not isinstance(args[0], (int, float)):
            reordered = []
            for idx in range(0, len(args), 2):
                reordered.append(args[idx + 1])
                reordered.append(args[idx])
        else:
            reordered = args
        return super(WalrusLedis, self).zadd(key, *reordered, **kwargs)

    def hash_exists(self, key):
        return self.execute_command('HKEYEXISTS', key)

    def __iter__(self):
        return self.scan()

    def scan(self, *args, **kwargs):
        return self._scan('XSCAN', *args, **kwargs)

    def _scan(self, cmd, match=None, count=None, ordering=None, limit=None):
        parts = ['KV', '']
        if match:
            parts.extend([Token('MATCH'), match])
        if count:
            parts.extend([Token('COUNT'), count])
        if ordering:
            parts.append(Token(ordering.upper()))
        return self._execute_scan(self, cmd, parts, limit)

    def update(self, values):
        return self.mset(values)

    def BitSet(self, key):
        return LedisBitSet(self, key)

    def Hash(self, key):
        return LedisHash(self, key)

    def List(self, key):
        return LedisList(self, key)

    def Set(self, key):
        return LedisSet(self, key)

    def ZSet(self, key):
        return LedisZSet(self, key)


class TestWalrusLedis(TestHelper, unittest.TestCase):
    def setUp(self):
        self.db = WalrusLedis()
        self.db.flushall()

    def test_scan(self):
        values = {
            'k1': 'v1',
            'k2': 'v2',
            'k3': 'v3',
            'charlie': 31,
            'mickey': 7,
            'huey': 5}
        self.db.update(values)
        results = self.db.scan()
        expected = ['charlie', 'huey', 'k1', 'k2', 'k3', 'mickey']
        self.assertEqual(list(results), expected)
        self.assertEqual([item for item in self.db], expected)

    def test_hash_iter(self):
        h = self.db.Hash('h_obj')
        h.clear()
        h.update({'k1': 'v1', 'k2': 'v2', 'k3': 'v3'})

        items = [item for item in h]
        self.assertEqual(items, ['k1', 'v1', 'k2', 'v2', 'k3', 'v3'])
        items = [item for item in h.scan(limit=2)]
        self.assertEqual(items, ['k1', 'v1', 'k2', 'v2'])

    def test_no_setitem_list(self):
        l = self.db.List('l_obj').clear()
        l.append('foo')
        self.assertRaises(TypeError, lambda: l.__setitem__(0, 'xx'))

    def test_set_iter(self):
        s = self.db.Set('s_obj').clear()
        s.add('charlie', 'huey', 'mickey')

        items = [item for item in s]
        self.assertEqual(sorted(items), ['charlie', 'huey', 'mickey'])
        items = [item for item in s.scan(limit=2, ordering='DESC')]
        self.assertEqual(items, ['mickey', 'huey'])

    def test_zset_iter(self):
        zs = self.db.ZSet('z_obj').clear()
        zs.add('zaizee', 3, 'mickey', 6, 'charlie', 31, 'huey', 3, 'nuggie', 0)

        items = [item for item in zs]
        self.assertEqual(items, [
            'charlie', '31',
            'huey', '3',
            'mickey', '6',
            'nuggie', '0',
            'zaizee', '3',
        ])

        items = [item for item in zs.scan(limit=3, ordering='DESC')]
        self.assertEqual(items, [
            'zaizee', '3',
            'nuggie', '0',
            'mickey', '6',
        ])

    def test_bit_set(self):
        b = self.db.BitSet('bitset_obj')
        b.clear()
        b[0] = 1
        b[1] = 1
        b[2] = 0
        b[3] = 1
        self.assertEqual(self.db[b.key], '\xd0')

        b[4] = 1
        self.assertEqual(self.db[b.key], '\xd8')
        self.assertEqual(b[0], 1)
        self.assertEqual(b[2], 0)

        self.db['b1'] = 'foobar'
        self.db['b2'] = 'abcdef'
        b = self.db.BitSet('b1')
        b2 = self.db.BitSet('b2')
        b &= b2
        self.assertEqual(self.db[b.key], '`bc`ab')
        self.assertEqual(str(b), '`bc`ab')

        self.db['b1'] = '\x00\xff\xf0'
        self.assertEqual(b.pos(1, 0), 8)
        self.assertEqual(b.pos(1, 2), 16)

        self.db['b1'] = '\x00\x00\x00'
        self.assertEqual(b.pos(1), -1)

    def test_sorting(self):
        items = ['charlie', 'zaizee', 'mickey', 'huey']
        sorted_items = sorted(items)

        l = self.db.List('l_obj').clear()
        l.extend(items)
        results = l.sort()
        self.assertEqual(results, sorted_items)

        dest = self.db.List('l_dest')
        l.sort(ordering='DESC', limit=3, store=dest.key)
        results = list(dest)
        self.assertEqual(results, ['zaizee', 'mickey', 'huey'])

        s = self.db.Set('s_obj').clear()
        s.add(*items)
        results = s.sort()
        self.assertEqual(results, sorted_items)

        results = s.sort(ordering='DESC', limit=3)
        self.assertEqual(results, ['zaizee', 'mickey', 'huey'])

        z = self.db.ZSet('z_obj').clear()
        z.add('charlie', 10, 'zaizee', 10, 'mickey', 3, 'huey', 4)
        results = z.sort()
        self.assertEqual(results, sorted_items)

        results = z.sort(ordering='DESC', limit=3)
        self.assertEqual(results, ['zaizee', 'mickey', 'huey'])


if __name__ == '__main__':
    unittest.main(argv=sys.argv)



import sys
import unittest

from vedis import Vedis

from walrus import *


class VedisList(List):
    def extend(self, value):
        return self.database.lmpush(self.key, value)

    def pop(self):
        return self.database.lpop(self.key)


class WalrusVedis(Vedis, Database):
    def __init__(self, filename=':mem:'):
        self._filename = filename
        Vedis.__init__(self, filename)

    def __repr__(self):
        if self._filename in (':memory:', ':mem:'):
            db_file = 'in-memory database'
        else:
            db_file = self._filename
        return '<WalrusVedis: %s>' % db_file

    def execute_command(self, *args, **options):
        raise ValueError('"%s" is not supported by Vedis.' % args[0])

    def parse_response(self, *args, **kwargs):
        raise RuntimeError('Error, parse_response should not be called.')

    def command(self, command_name):
        return self.register(command_name)

    # Compatibility with method names from redis-py.
    def getset(self, key, value):
        return self.get_set(key, value)

    def incrby(self, name, amount=1):
        return self.incr_by(name, amount)

    def decrby(self, name, amount=1):
        return self.decr_by(name, amount)

    # Compatibility with method signatures.
    def mset(self, **data):
        return super(WalrusVedis, self).mset(data)

    def mget(self, *keys):
        return super(WalrusVedis, self).mget(list(keys))

    def __getitem__(self, key):
        try:
            return super(WalrusVedis, self).__getitem__(key)
        except KeyError:
            pass

    def sadd(self, key, *items):
        return super(WalrusVedis, self).smadd(key, list(items))

    # Override the container types since Vedis provides its own using the
    # same method-names as Walrus, and we want the Walrus containers.
    def Hash(self, key):
        return Hash(self, key)

    def Set(self, key):
        return Set(self, key)

    def List(self, key):
        return VedisList(self, key)

    def not_supported(name):
        def decorator(self, *args, **kwargs):
            raise ValueError('%s is not supported by Vedis.' % name)
        return decorator

    ZSet = not_supported('ZSet')
    Array = not_supported('Array')
    HyperLogLog = not_supported('HyperLogLog')
    pipeline = not_supported('pipeline')
    lock = not_supported('lock')
    pubsub = not_supported('pubsub')


class TestWalrusVedis(unittest.TestCase):
    def setUp(self):
        self.db = WalrusVedis()

    def test_basic(self):
        self.db['foo'] = 'bar'
        self.assertEqual(self.db['foo'], 'bar')
        self.assertTrue('foo' in self.db)
        self.assertFalse('xx' in self.db)
        self.assertIsNone(self.db['xx'])

        self.db.mset(k1='v1', k2='v2', k3='v3')
        results = self.db.mget('k1', 'k2', 'k3', 'kx')
        self.assertEqual(list(results), ['v1', 'v2', 'v3', None])

        self.db.append('foo', 'baz')
        self.assertEqual(self.db.get('foo'), 'barbaz')

        self.db.incr_by('counter', 1)
        self.assertEqual(self.db.incr_by('counter', 5), 6)
        self.assertEqual(self.db.decr_by('counter', 2), 4)

        self.assertEqual(self.db.strlen('foo'), 6)
        self.assertEqual(self.db.getset('foo', 'nug'), 'barbaz')
        self.assertEqual(self.db['foo'], 'nug')

        self.assertFalse(self.db.setnx('foo', 'xxx'))
        self.assertTrue(self.db.setnx('bar', 'yyy'))
        self.assertEqual(self.db['bar'], 'yyy')

        del self.db['foo']
        self.assertFalse('foo' in self.db)

    def test_hash(self):
        h = self.db.Hash('hash_obj')
        h['k1'] = 'v1'
        h.update({'k2': 'v2', 'k3': 'v3'})
        self.assertEqual(h.as_dict(), {'k1': 'v1', 'k2': 'v2', 'k3': 'v3'})

        self.assertEqual(h['k2'], 'v2')
        self.assertIsNone(h['kx'])
        self.assertTrue('k2' in h)
        self.assertEqual(len(h), 3)

        del h['k2']
        del h['kxx']
        self.assertEqual(sorted(h.keys()), ['k1', 'k3'])
        self.assertEqual(sorted(h.values()), ['v1', 'v3'])

    def test_list(self):
        l = self.db.List('list_obj')
        l.prepend('charlie')
        l.extend(['mickey', 'huey', 'zaizee'])
        self.assertEqual(l[0], 'charlie')
        self.assertEqual(l[-1], 'zaizee')
        self.assertEqual(len(l), 4)
        self.assertEqual(l.pop(), 'charlie')

    def test_set(self):
        s = self.db.Set('set_obj')
        s.add('charlie')
        s.add('charlie', 'huey', 'mickey')
        self.assertEqual(len(s), 3)
        self.assertTrue('huey' in s)
        self.assertFalse('xx' in s)
        del s['huey']
        self.assertFalse('huey' in s)
        self.assertEqual(s.members(), set(['charlie', 'mickey']))

        s1 = self.db.Set('s1')
        s2 = self.db.Set('s2')
        s1.add(*map(str, range(5)))
        s2.add(*map(str, range(3, 7)))
        self.assertEqual(s1 - s2, set(['0', '1', '2']))
        self.assertEqual(s2 - s1, set(['5', '6']))
        self.assertEqual(s1 & s2, set(['3', '4']))

    def test_unsupported(self):
        def assertUnsupported(cmd, *args):
            method = getattr(self.db, cmd)
            self.assertRaises(ValueError, method, *args)

        # Just check a handful of methods.
        assertUnsupported('zadd', 'zs', 'foo', 1)
        assertUnsupported('ZSet', 'zs')
        assertUnsupported('rpush', 'l_obj', 'val')
        assertUnsupported('rpop', 'l_obj')
        assertUnsupported('ltrim', 'l_obj', 0, 1)
        assertUnsupported('lrem', 'l_obj', 3, 1)

    def test_custom_commands(self):
        @self.db.command('KTITLE')
        def _ktitle_impl(context, key):
            value = context[key]
            if value:
                context[key] = value.title()
                return True
            return False

        self.db['n1'] = 'charlie'
        self.db['n2'] = 'huey'
        self.assertTrue(_ktitle_impl('n1'))
        self.assertEqual(self.db['n1'], 'Charlie')

        self.assertTrue(self.db.execute('KTITLE n2'))
        self.assertEqual(self.db['n2'], 'Huey')

        self.assertFalse(self.db.execute('KTITLE nx'))
        self.assertIsNone(self.db['nx'])


if __name__ == '__main__':
    unittest.main(argv=sys.argv)
